from box import Box
from scipy.spatial.transform import Rotation

import opengate_core as g4
from .base import SourceBase
from .utility import (
    all_beta_plus_radionuclides,
    get_spectrum,
    compute_cdf_and_total_yield,
)
from ..actors.biasingactors import generic_source_default_aa
from ..base import process_cls
from ..utility import g4_units
from ..exception import fatal, warning


def _generic_source_default_position():
    """
    types {"sphere", "point", "box", "disc", "cylinder"};
    """
    return Box(
        {
            "type": "point",
            "radius": 0,
            "sigma_x": 0,
            "sigma_y": 0,
            "size": [0, 0, 0],
            "translation": [0, 0, 0],
            "rotation": Rotation.identity().as_matrix(),
            "confine": None,
        }
    )


def _generic_source_default_direction():
    return Box(
        {
            "type": "iso",
            "theta": [0, 180 * g4_units.deg],
            "phi": [0, 360 * g4_units.deg],
            "momentum": [0, 0, 1],
            "focus_point": [0, 0, 0],
            "sigma": [0, 0],
            "acceptance_angle": generic_source_default_aa(),
            "accolinearity_flag": False,
            "accolinearity_fwhm": 0.5 * g4_units.deg,
            "histogram_theta_weights": [],
            "histogram_theta_angles": [],
            "histogram_phi_weights": [],
            "histogram_phi_angles": [],
        }
    )


def _generic_source_default_energy():
    return Box(
        {
            "type": "mono",
            "mono": 0,
            "sigma_gauss": 0,
            "is_cdf": False,
            "min_energy": None,
            "max_energy": None,
            "spectrum_type": None,
            "spectrum_weights": [],
            "spectrum_energies": [],
            "spectrum_energy_bin_edges": [],
            "spectrum_histogram_interpolation": None,
        }
    )


def _setter_hook_generic_source_particle(self, particle):
    # particle must be a str
    if not isinstance(particle, str):
        fatal(f"the .particle user info must be a str, while it is {type(str)}")
    # if it does not start with ion, we consider this is a simple particle (gamma, e+ etc)
    if not particle.startswith("ion"):
        return particle
    # if start with ion, it is like 'ion 9 18' with Z A E
    words = particle.split(" ")
    if len(words) > 1:
        self.ion.Z = int(words[1])
    if len(words) > 2:
        self.ion.A = int(words[2])
    if len(words) > 3:
        self.ion.E = int(words[3])
    return particle


class GenericSource(SourceBase, g4.GateGenericSource):
    """
    GenericSource close to the G4 SPS, but a bit simpler.
    The G4 source created by this class is GateGenericSource.
    """

    # hints for IDE
    particle: str
    ion: Box
    weight: float
    weight_sigma: float
    user_particle_life_time: float
    tac_times: list
    tac_activities: list
    direction_relative_to_attached_volume: bool
    position: Box
    direction: Box
    energy: Box

    user_info_defaults = {
        "particle": (
            "gamma",
            {
                "doc": "Name of the particle generated by the source (gamma, e+ ... or an ion such as 'ion 9 18')",
                "setter_hook": _setter_hook_generic_source_particle,
            },
        ),
        "ion": (
            Box({"Z": 0, "A": 0, "E": 0}),
            {
                "doc": "If the particle is an ion, you must set Z: Atomic Number, A: Atomic Mass (nn + np +nlambda), E: Excitation energy (i.e. for metastable)"
            },
        ),
        "weight": (
            -1,
            {"doc": "Particle initial weight (for variance reduction technique)"},
        ),
        "weight_sigma": (
            -1,
            {
                "doc": "if not negative, the weights of the particle are a Gaussian distribution with this sigma"
            },
        ),
        "user_particle_life_time": (
            -1,
            {"doc": "FIXME "},
        ),
        "tac_times": (
            None,
            {
                "doc": "TAC: Time Activity Curve, this set the vector for the times. Must be used with tac_activities."
            },
        ),
        "tac_activities": (
            None,
            {
                "doc": "TAC: Time Activity Curve, this set the vector for the activities. Must be used with tac_times."
            },
        ),
        "direction_relative_to_attached_volume": (
            False,
            {
                "doc": "Should we update the direction of the particle "
                "when the volume is moved (with dynamic parametrisation)?"
            },
        ),
        "position": (
            _generic_source_default_position(),
            {"doc": "Define the position of the primary particles"},
        ),
        "direction": (
            _generic_source_default_direction(),
            {"doc": "Define the direction of the primary particles"},
        ),
        "energy": (
            _generic_source_default_energy(),
            {"doc": "Define the energy of the primary particles"},
        ),
        "polarization": (
            [],
            {"doc": "Polarization of the particle (3 Stokes parameters)."},
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.__initcpp__()
        self.total_zero_events = 0
        self.total_skipped_events = 0
        if not self.user_info.particle.startswith("ion"):
            return
        words = self.user_info.particle.split(" ")
        if len(words) > 1:
            self.user_info.ion.Z = words[1]
        if len(words) > 2:
            self.user_info.ion.A = words[2]
        if len(words) > 3:
            self.user_info.ion.E = words[3]

    def __initcpp__(self):
        g4.GateGenericSource.__init__(self)

    def initialize(self, run_timing_intervals):
        if not isinstance(self.position, Box):
            fatal(
                f"Generic Source: user_info.position must be a Box, but is: {self.position}"
            )
        if not isinstance(self.direction, Box):
            fatal(
                f"Generic Source: user_info.direction must be a Box, but is: {self.direction}"
            )
        if not isinstance(self.energy, Box):
            fatal(
                f"Generic Source: user_info.energy must be a Box, but is: {self.energy}"
            )

        if self.particle == "back_to_back":
            # force the energy to 511 keV
            self.energy.type = "mono"
            self.energy.mono = 511 * g4_units.keV

        # check the energy type
        l = [
            "mono",
            "gauss",
            "F18_analytic",
            "O15_analytic",
            "C11_analytic",
            "histogram",
            "spectrum_discrete",
            "spectrum_histogram",
            "range",
        ]
        l.extend(all_beta_plus_radionuclides)
        if self.energy.type not in l:
            fatal(
                f"Cannot find the energy type {self.energy.type} for the source {self.name}.\n"
                f"Available types are {l}"
            )

        # check if the energy spectrum type if known
        valid_spectrum_types = [
            "discrete",
            "histogram",
            "interpolated",
        ]
        if self.energy.spectrum_type is not None:
            if self.energy.spectrum_type not in valid_spectrum_types:
                fatal(
                    f"Cannot find the energy spectrum type {self.energy.spectrum_type} for the source {self.name}.\n"
                    f"Available types are {valid_spectrum_types}"
                )

        # special case for beta plus energy spectra
        # FIXME put this elsewhere
        if self.particle == "e+":
            if self.energy.type in all_beta_plus_radionuclides:
                data = get_spectrum(self.user_info.energy.type, "e+", "radar")
                ene = data[:, 0] / 1000  # convert from KeV to MeV
                proba = data[:, 1]
                cdf, _ = compute_cdf_and_total_yield(proba, ene)
                # total = total * 1000  # (because was in MeV)
                # self.user_info.activity *= total
                self.energy.is_cdf = True
                self.SetEnergyCDF(ene)
                self.SetProbabilityCDF(cdf)

        self.update_tac_activity()

        # histogram parameters: histogram_weight, histogram_energy"
        ene = self.energy
        if ene.type == "histogram":
            if len(ene.histogram_weight) != len(ene.histogram_energy):
                fatal(
                    f'For the source {self.name}, the parameters "energy", '
                    f'"histogram_energy" and "histogram_weight" must have the same length'
                )

        # check direction type
        l = ["iso", "histogram", "momentum", "focused", "beam2d"]
        if self.direction.type not in l:
            fatal(
                f"Cannot find the direction type {self.direction.type} for the source {self.name}.\n"
                f"Available types are {l}"
            )

        # logic for half life and user_particle_life_time
        if self.half_life > 0:
            # if the user set the half life and not the user_particle_life_time
            # we force the latter to zero
            if self.user_particle_life_time < 0:
                self.user_particle_life_time = 0

        # initialize
        SourceBase.initialize(self, run_timing_intervals)

        if self.n > 0 and self.activity > 0:
            fatal(
                f"Cannot use both the two parameters 'n' and 'activity' at the same time. "
            )
        if self.n == 0 and self.activity == 0:
            fatal(f"You must set one of the two parameters 'n' or 'activity'.")
        if self.activity > 0:
            self.n = 0
        if self.n > 0:
            self.activity = 0
        # warning for non-used ?

        # check confine
        if self.position.confine:
            if self.position.type == "point":
                warning(
                    f"In source {self.name}, "
                    f"confine is used, while position.type is point ... really ?"
                )

    def check_ui_activity(self, ui):
        # FIXME: This should rather be a function than a method
        # FIXME: self actually holds the parameters n and activity, but the ones from ui are used here.
        if ui.n > 0 and ui.activity > 0:
            fatal(f"Cannot use both n and activity, choose one: {self.user_info}")
        if ui.n == 0 and ui.activity == 0:
            fatal(f"Choose either n or activity : {self.user_info}")
        if ui.activity > 0:
            ui.n = 0
        if ui.n > 0:
            ui.activity = 0

    def check_confine(self, ui):
        # FIXME: This should rather be a function than a method
        # FIXME: self actually holds the parameters n and activity, but the ones from ui are used here.
        if ui.position.confine:
            if ui.position.type == "point":
                warning(
                    f"In source {ui.name}, "
                    f"confine is used, while position.type is point ... really ?"
                )

    def prepare_output(self):
        SourceBase.prepare_output(self)
        # store the output from G4 object
        self.total_zero_events = self.GetTotalZeroEvents()
        self.total_skipped_events = self.GetTotalSkippedEvents()

    def update_tac_activity(self):
        if self.tac_times is None and self.tac_activities is None:
            return
        if len(self.tac_times) != len(self.tac_activities):
            fatal(
                f"option tac_activities must have the same size as tac_times in source '{self.name}'"
            )
        # it is important to set the starting time for this source as the tac
        # may start later than the simulation timing
        self.start_time = self.tac_times[0]
        self.activity = self.tac_activities[0]
        self.SetTAC(self.tac_times, self.tac_activities)

    def can_predict_number_of_events(self):
        aa = self.direction.acceptance_angle
        if aa.intersection_flag or aa.normal_flag:
            if aa.skip_policy == "ZeroEnergy":
                return True
            return False
        return True


process_cls(GenericSource)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Context: In Shibuya et al. 2007 [1], it was shown that acollinearity of annihilation
photons in a human subject follows a double Gaussian distribution with a combined FWHM
of 0.55 deg.
While the double Gaussian distribution currently cannot be reproduced in GATE, setting the MeanEnergyPerIonPair of the material to 6.0 eV results in a 2D Gaussian with a FWHM of 0.55 deg.

Note: Changing the material to "Body" does not change the value of acollinearity
compared to water. Unknown if it is a limitation in the simulation or if it is due to
0.5 deg being obtained in older setup with water at 20 deg vs human that are a little
more warm?

[1] https://iopscience.iop.org/article/10.1088/0031-9155/52/17/010
"""

from test079_acollin_helpers import *
import opengate.tests.utility as tu
import matplotlib.pyplot as plt


#########################################################################################
# Simulations configuration that may be relevant to change
#########################################################################################
# Mean energy of Ion Pair to use. 6.0 eV seems to results in 0.55 deg FWHM
mean_energy = 6.0 * eV
# Key added to output to make sure that multi-threading the tests does not backfire
test_key = "p5"


#########################################################################################
# Main : We use this to launch the test
#########################################################################################
if __name__ == "__main__":
    paths = tu.get_default_test_paths(__file__, output_folder="test079")

    # Define core of the simulation, including physics
    sim = setup_simulation_engine(paths)

    # add a waterbox
    wb = sim.add_volume("Box", "waterbox")
    wb.size = [50 * cm, 50 * cm, 50 * cm]
    wb.material = "G4_WATER"

    # test mat properties
    mat = sim.volume_manager.find_or_build_material(wb.material)
    ionisation = mat.GetIonisation()
    print(
        f"material {wb.material} mean excitation energy is {ionisation.GetMeanExcitationEnergy() / eV} eV"
    )
    print(
        f"material {wb.material} mean energy per ion pair is {ionisation.GetMeanEnergyPerIonPair() / eV} eV"
    )

    # set the source
    source = sim.add_source("GenericSource", "f18")
    source.particle = "ion 9 18"
    source.energy.mono = 0
    source.activity = 10000 * Bq
    source.direction.type = "iso"

    # add phase actor
    phsp = setup_actor(sim, "phsp", wb.name)
    phsp.output_filename = (
        paths.output / f"annihilation_photons_with_mepip_{test_key}.root"
    )
    ionisation.SetMeanEnergyPerIonPair(mean_energy)
    print(f"set MeanEnergyPerIonPair to {ionisation.GetMeanEnergyPerIonPair() / eV} eV")

    # go
    sim.run()

    # test: with mean energy, acolinearity amplitude should have a Rayleigh distribution
    gamma_pairs = read_gamma_pairs(phsp.output_filename)
    acollinearity_angles = compute_acollinearity_angles(gamma_pairs)

    acolin_scale = plot_acolin_case_mepip(mean_energy, acollinearity_angles)

    f = paths.output / f"acollinearity_angles_{test_key}.png"
    plt.savefig(f)
    print(f"Plot was saved in {f}")

    # final
    # Basic acolin
    is_ok_p2 = np.isclose(acolin_scale * 2.355, 0.55, atol=0.2)

    tu.test_ok(is_ok_p2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test036_adder_depth_helpers as t036
from opengate.tests import utility
from opengate import g4_units

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test036_adder_depth", "test036"
    )

    # create and run the simulation
    mm = g4_units.mm
    sim = t036.create_simulation("param", paths, "_par1")

    # sim.visu = True
    sim.visu_type = "qt"

    # remove one head
    head = sim.volume_manager.get_volume("SPECThead")
    head.translation = head.translation[0]
    head.rotation = head.rotation[0]

    # enlarge the source
    source = sim.source_manager.get_source("src2")
    source.position.radius = 150 * mm

    if sim.visu:
        source = sim.source_manager.get_source("src2")
        source.activity = source.activity / 1000
        source = sim.source_manager.get_source("src1")
        source.activity = source.activity / 1000

    # add a proj actor
    proj = sim.add_actor("DigitizerProjectionActor", "proj")
    proj.attached_to = "crystal"
    fname = "proj1.mha"
    proj.output_filename = fname.replace(".mha", "-1.mha")
    proj.size = [128, 128]
    proj.spacing = [5 * mm, 5 * mm]

    # start simulation
    sim.run()

    # test the output
    stats = sim.get_actor("Stats")
    is_ok = utility.assert_images(
        paths.output_ref / fname,
        proj.get_output_path("counts"),
        stats,
        tolerance=38,
        ignore_value_data2=0,
        axis="y",
        fig_name=paths.output / f"proj.png",
        sum_tolerance=1.5,
    )
    utility.print_test(is_ok, f"Compare image proj:")
    utility.test_ok(is_ok)



import opengate_core as g4
from .base import ActorBase
from ..base import process_cls
from box import Box
from ..utility import g4_units
from .actoroutput import ActorOutputBase
import numpy as np


def generic_source_default_aa():
    # aa = Angular Acceptance
    # this is used to control the direction of events in
    # the generic source, but is also used in the SplitComptonActor
    return Box(
        {
            "skip_policy": "SkipEvents",
            "max_rejection": 10000,
            "volumes": [],
            "intersection_flag": False,
            "normal_flag": False,
            "forced_direction_flag": False,
            "normal_vector": [0, 0, 1],
            "normal_tolerance": 3 * g4_units.deg,
            "normal_tolerance_min_distance": 0 * g4_units.cm,
            "distance_dependent_normal_tolerance": False,
            "angle1": 90 * g4_units.degree,
            "distance1": 0 * g4_units.cm,
            "angle2": 20 * g4_units.degree,
            "distance2": 50 * g4_units.cm,
        }
    )


def distance_dependent_angle_tolerance(a1, a2, d1, d2, dist):
    a = (1 / np.tan(a1) - 1 / np.tan(a2)) / (d1 - d2)
    b = 1 / np.tan(a1) - a * d1
    tol = np.arctan(1.0 / (a * dist + b))
    if tol < 0:
        tol = 90 * g4_units.deg
    return tol


def _setter_hook_particles(self, value):
    if isinstance(value, str):
        return [value]
    else:
        return list(value)


class GenericBiasingActorBase(ActorBase):
    """
    Actors based on the G4GenericBiasing class of GEANT4. This class provides tools to interact with GEANT4 processes
    during a simulation, allowing direct modification of process properties.
    """

    # hints for IDE
    bias_primary_only: bool
    bias_only_once: bool
    particles: list

    user_info_defaults = {
        "bias_primary_only": (
            True,
            {
                "doc": "If true, the biasing mechanism is applied only to particles with a ParentID of 1",
            },
        ),
        "bias_only_once": (
            True,
            {
                "doc": "If true, the biasing mechanism is applied only once per particle history",
            },
        ),
        "particles": (
            [
                "all",
            ],
            {
                "doc": "Specifies the particles to split. The default value, all, includes all particles",
                "setter_hook": _setter_hook_particles,
            },
        ),
        "unbiased_volumes": (
            [],
            {
                "doc": "A list of volumes where this actor's biasing is disabled, allowing particles to be tracked with normal, unbiased physics. ",
            },
        ),
        "minimal_weight": (
            -1,
            {
                "doc": "if the particle weight become lower than this value, the particle is killed. "
                "Negative values are considered with minimal_weight = min double default (1e-300)"
            },
        ),
    }


class SplitProcessActorBase(GenericBiasingActorBase):
    """
    This actor  enables non-physics-based particle splitting (e.g., pure geometrical splitting) to introduce biasing
    into simulations. SplittingActorBase serves as a foundational class for particle splitting operations,
    with parameters for configuring the splitting behavior based on various conditions.
    """

    # hints for IDE
    splitting_factor: int

    user_info_defaults = {
        "splitting_factor": (
            1,
            {
                "doc": "Specifies the number of particles to generate each time the splitting mechanism is applied",
            },
        ),
    }


class BremsstrahlungSplittingActor(
    SplitProcessActorBase, g4.GateBremsstrahlungSplittingOptrActor
):
    """
    This splitting actor enables process-based splitting specifically for bremsstrahlung process. Each time a Brem
    process occurs, its behavior is modified by generating multiple secondary Brem scattering tracks
    (splitting factor) attached to  the initial charged particle.

    This actor is not really needed as Geant4 already propose this with:
    /process/em/setSecBiasing eBrem my_region 100 50 MeV
    But we use it as a test/example.

    """

    # hints for IDE
    processes: list

    user_info_defaults = {
        "processes": (
            ["eBrem"],
            {
                "doc": "Specifies the process split by this actor. This parameter is set to eBrem, as the actor is specifically developed for this process. It is recommended not to modify this setting.",
            },
        ),
    }

    processes = ("eBrem",)

    def __init__(self, *args, **kwargs):
        SplitProcessActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateBremsstrahlungSplittingOptrActor.__init__(self, {"name": self.name})

    def initialize(self):
        SplitProcessActorBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()


class GammaFreeFlightActor(GenericBiasingActorBase, g4.GateGammaFreeFlightOptrActor):
    """

    Warning: as a G4VBiasingOperator, the attachTo operation MUST be done
    1) before the StartSimulationAction and 2) for each thread
    So it is set in cpp Configure and ConfigureForWorker
    that is specifically called in engines.py
    in the register_sensitive_detectors function

    Also, PreUserTrackingAction is needed because StartTracking is not used in MT.
    """

    # hints for IDE
    processes: list

    # this biased actor DO NOT work for GammaGeneralProc
    processes = ["compt", "phot", "conv", "Rayl"]
    particles = ["gamma"]

    def __init__(self, *args, **kwargs):
        GenericBiasingActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateGammaFreeFlightOptrActor.__init__(self, {"name": self.name})

    def initialize(self):
        GenericBiasingActorBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        g4.GateGammaFreeFlightOptrActor.StartSimulationAction(self)


class ActorOutputScatterSplittingFreeFlightActor(ActorOutputBase):
    """
    Some output statistics computed during ScatterSplittingFreeFlightActor
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # self.splitting_factor = kwargs.get("belongs_to").user_info.splitting_factor

        # predefine the split_info
        self.split_info = Box()
        self.split_info.nb_tracks = 0
        self.split_info.nb_tracks_with_free_flight = 0
        self.split_info.nb_compt_splits = 0
        self.split_info.nb_rayl_splits = 0
        self.split_info.nb_compt_tracks = 0
        self.split_info.nb_rayl_tracks = 0
        self.split_info.nb_killed_non_gamma_particles = 0
        self.split_info.nb_killed_gammas_compton_level = 0
        self.split_info.nb_killed_gammas_exiting = 0
        self.split_info.compton_splitting_factor = 1
        self.split_info.rayleigh_splitting_factor = 1

    def get_processed_output(
        self, infos, compton_splitting_factor, rayleigh_splitting_factor
    ):
        self.split_info.nb_tracks = infos["nb_tracks"]
        self.split_info.nb_tracks_with_free_flight = infos["nb_tracks_with_free_flight"]
        self.split_info.nb_compt_splits = infos["nb_compt_splits"]
        self.split_info.nb_rayl_splits = infos["nb_rayl_splits"]
        self.split_info.nb_compt_tracks = infos["nb_compt_tracks"]
        self.split_info.nb_rayl_tracks = infos["nb_rayl_tracks"]
        self.split_info.nb_killed_non_gamma_particles = infos[
            "nb_killed_non_gamma_particles"
        ]
        self.split_info.nb_killed_gammas_compton_level = infos[
            "nb_killed_gammas_compton_level"
        ]
        self.split_info.nb_killed_gammas_exiting = infos["nb_killed_gammas_exiting"]
        self.split_info.compton_splitting_factor = compton_splitting_factor
        self.split_info.rayleigh_splitting_factor = rayleigh_splitting_factor
        return self.split_info

    def __str__(self):
        s = ""
        for key, value in self.split_info.items():
            s += f"{key}: {value}\n"

        if (
            self.split_info.compton_splitting_factor > 0
            and self.split_info.nb_compt_splits > 0
        ):
            f = self.split_info.nb_compt_tracks / (
                self.split_info.nb_compt_splits
                * self.split_info.compton_splitting_factor
            )
            s += f"Fraction of FF compton: {f*100:.2f} %\n"

        if (
            self.split_info.rayleigh_splitting_factor > 0
            and self.split_info.nb_rayl_splits > 0
        ):
            f = self.split_info.nb_rayl_tracks / (
                self.split_info.nb_rayl_splits
                * self.split_info.rayleigh_splitting_factor
            )
            s += f"Fraction of FF rayleigh: {f*100:.2f} %\n"

        if self.split_info.nb_compt_tracks < 1 and self.split_info.nb_rayl_tracks < 1:
            f = 0
        else:
            f = self.split_info.nb_tracks_with_free_flight / (
                self.split_info.nb_compt_tracks + self.split_info.nb_rayl_tracks
            )
        s += f"Check split vs ff (should be 100): {f*100:.2f} %\n"
        return s


class ScatterSplittingFreeFlightActor(
    SplitProcessActorBase, g4.GateScatterSplittingFreeFlightOptrActor
):
    """
    Split Compton process for gamma. The initial gamma is tracked until it goes out of the volume.
    Split gammas are tracked with free flight and Angular Acceptance
    """

    # hints for IDE
    processes: list

    # user info
    user_info_defaults = {
        "max_compton_level": (
            10,
            {
                "doc": "Compton are split until this max level is reached (then the initial gamma is killed).",
            },
        ),
        "compton_splitting_factor": (
            -1,
            {
                "doc": "All Compton interactions will be split by this factor (if -1, set by splitting_factor).",
            },
        ),
        "rayleigh_splitting_factor": (
            -1,
            {
                "doc": "All Rayleigh interactions will be split by this factor (if -1, set by splitting_factor).",
            },
        ),
        "acceptance_angle": (
            generic_source_default_aa(),
            {
                "doc": "See generic source",
            },
        ),
        "kill_interacting_in_volumes": (
            [],
            {
                "doc": "When a non-split particle enters one of those volumes, it is killed.",
            },
        ),
    }

    # Do NOT work with GammaGeneralProc
    # Do NOT include "phot" process here.
    processes = ["compt", "Rayl"]
    particles = ["gamma"]

    user_output_config = {
        "info": {
            "actor_output_class": ActorOutputScatterSplittingFreeFlightActor,
        },
    }

    def __init__(self, *args, **kwargs):
        SplitProcessActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateScatterSplittingFreeFlightOptrActor.__init__(self, {"name": self.name})

    def __str__(self):
        s = self.user_output["info"].__str__()
        return s

    def initialize(self):
        SplitProcessActorBase.initialize(self)
        if self.user_info.compton_splitting_factor == -1:
            self.user_info.compton_splitting_factor = self.user_info.splitting_factor
        if self.user_info.rayleigh_splitting_factor == -1:
            self.user_info.rayleigh_splitting_factor = self.user_info.splitting_factor
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def EndSimulationAction(self):
        g4.GateScatterSplittingFreeFlightOptrActor.EndSimulationAction(self)
        self.user_output["info"].get_processed_output(
            self.GetBiasInformation(),
            self.user_info.compton_splitting_factor,
            self.user_info.rayleigh_splitting_factor,
        )


process_cls(GenericBiasingActorBase)
process_cls(SplitProcessActorBase)
process_cls(BremsstrahlungSplittingActor)
process_cls(GammaFreeFlightActor)
process_cls(ActorOutputScatterSplittingFreeFlightActor)
process_cls(ScatterSplittingFreeFlightActor)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import pathlib

# This is just a dummy simulation to test the json archiving functionality
# It does not simulate any meaningful scenario

if __name__ == "__main__":
    pathFile = pathlib.Path(__file__).parent.resolve()
    paths = utility.get_default_test_paths(__file__)
    paths_test009 = utility.get_default_test_paths(__file__, "gate_test009_voxels")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False

    # store a json archive
    sim.store_json_archive = True
    sim.store_input_files = True
    sim.json_archive_filename = "simu_test069.json"
    sim.output_dir = paths.output / "test069"

    # add a material database
    sim.volume_manager.add_material_database(
        pathFile.parent / "data" / "GateMaterials.db"
    )

    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV

    #  change world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]

    # create a waterbox, but do not add it to the simulation
    waterbox = sim.volume_manager.create_volume("Box", "Waterbox")
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.material = "G4_WATER"

    # create and add a rod
    rod = sim.add_volume("Tubs", "rod")
    rod.rmax = 1 * cm
    rod.rmin = 0
    rod.dz = 6 * cm
    rod.mother = "world"
    rod.material = "G4_PLEXIGLASS"

    # "punch a hole" in the waterbox via boolean subtraction
    waterbox_with_hole = gate.geometry.volumes.subtract_volumes(
        waterbox, rod, new_name="waterbox_with_hole"
    )
    # and add the resulting volume
    sim.add_volume(waterbox_with_hole)

    # Note: the rod is higher (2 x 6 cm) than the waterbox so the rod sticks out of the punched-through waterbox

    # set production cuts in the rod volume
    sim.physics_manager.set_production_cut(
        volume_name=rod.name, value=1.78 * gate.g4_units.MeV, particle_name="proton"
    )

    # add an image volume
    patient = sim.add_volume("Image", "patient")
    patient.image = paths.data / "patient-4mm.mhd"
    patient.mother = "world"
    patient.translation = [0, 0, 30 * cm]
    patient.material = "G4_AIR"  # material used by default
    patient.voxel_materials = [
        [-2000, -900, "G4_AIR"],
        [-900, -100, "Lung"],
        [-100, 0, "G4_ADIPOSE_TISSUE_ICRP"],
        [0, 300, "G4_TISSUE_SOFT_ICRP"],
        [300, 800, "G4_B-100_BONE"],
        [800, 6000, "G4_BONE_COMPACT_ICRU"],
    ]

    stat_actor = sim.add_actor("SimulationStatisticsActor", name="stat_actor")

    # add a source so that this simulation can run
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 230 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 1 * mm
    source.position.translation = [0, 0, -80 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 10

    # run
    sim.run()

    # test the file content -> NO, there are some abs filenames ...
    fn1 = paths.output / "test069" / sim.json_archive_filename
    print(fn1)
    """fn2 = paths.output_ref / "test069" / sim.json_archive_filename
    f1 = open(fn1)
    j1 = json.load(f1)
    f2 = open(fn2)
    j2 = json.load(f2)
    is_ok = j1 == j2
    print(fn1)
    print(fn2)
    utility.print_test(is_ok, f"Compare json gate output with reference")"""
    is_ok = fn1.exists()

    utility.test_ok(is_ok)



import opengate_core
from opengate.contrib.tps.ionbeamtherapy import *
from ...exception import fatal


class TreatmentPlanPhsSource(TreatmentPlanSource):

    def __init__(self, name, sim):
        super().__init__(name, sim)
        self.name = name
        self.rotation = Rotation.identity()
        self.translation = [0, 0, 0]
        self.spots = None
        self.phaseSpaceFolder = ""
        self.phaseSpaceList_file_name = ""
        self.phaseSpaceList = {}
        self.position_key_x = "PrePositionLocal_X"
        self.position_key_y = "PrePositionLocal_Y"
        self.position_key_z = "PrePositionLocal_Z"
        self.direction_key_x = "PreDirectionLocal_X"
        self.direction_key_y = "PreDirectionLocal_Y"
        self.direction_key_z = "PreDirectionLocal_Z"
        self.rotate_PhS_Source = Rotation.identity()
        self.energy_key = "KineticEnergy"
        self.weight_key = "Weight"
        self.PDGCode_key = "PDGCode"
        self.generate_until_next_primary = False
        self.primary_lower_energy_threshold = 0
        self.primary_PDGCode = 0
        self.n_sim = 0
        self.sim = sim  # simulation obj to which we want to add the tpPhS source
        self.distance_source_to_isocenter = None
        # SMX to Isocenter distance
        self.distance_stearmag_to_isocenter_x = None
        # SMY to Isocenter distance
        self.distance_stearmag_to_isocenter_y = None
        self.batch_size = None
        self.entry_start = None

    def __del__(self):
        pass

    def set_distance_source_to_isocenter(self, distance):
        self.distance_source_to_isocenter = distance
        self.d_nozzle_to_iso = distance

    def set_distance_stearmag_to_isocenter(self, distance_x, distance_y):
        self.d_stearMag_to_iso_x = distance_x
        self.d_stearMag_to_iso_y = distance_y
        self.distance_stearmag_to_isocenter_x = distance_x
        self.distance_stearmag_to_isocenter_y = distance_y

    def set_phaseSpaceList_file_name(self, file_name):
        self.phaseSpaceList_file_name = str(file_name)

    def set_phaseSpaceFolder(self, folder_name):
        self.phaseSpaceFolder = str(folder_name)

    def initialize_tpPhssource(self):
        if self.batch_size is None:
            self.batch_size = 30000
        # verify that all necessary parameters are set
        self.verify_necessary_parameters_are_set()
        # read in the phase space list
        self.phaseSpaceList = self.read_list_of_Phs(
            self.phaseSpaceList_file_name, self.phaseSpaceFolder
        )
        # verify the phase space list
        self.verify_phs_files_exist(self.phaseSpaceList)
        spots_array = self.spots
        sim = self.sim
        nSim = self.n_sim

        # mapping factors between iso center plane and nozzle plane (due to steering magnets)
        cal_proportion_factor = lambda d_magnet_iso: (
            1
            if (d_magnet_iso == float("inf"))
            else (d_magnet_iso - self.d_nozzle_to_iso) / d_magnet_iso
        )
        self.proportion_factor_x = cal_proportion_factor(self.d_stearMag_to_iso_x)
        self.proportion_factor_y = cal_proportion_factor(self.d_stearMag_to_iso_y)

        tot_sim_particles = 0
        # initialize a pencil beam for each spot
        for i, spot in enumerate(spots_array):
            # simulate a fraction of the beam particles for this spot
            nspot = np.round(spot.beamFraction * nSim)
            if nspot == 0:
                continue
            tot_sim_particles += nspot
            # create a source
            source = sim.add_source("PhaseSpaceSource", f"{self.name}_spot_{i}")

            # set energy
            # find corresponding phase space file
            if self.phaseSpaceList.get(spot.energy) is not None:
                source.phsp_file = self.phaseSpaceList.get(spot.energy)
            else:
                fatal(
                    f"Error in TreatmentPlanPhsSource: Energy requested from plan file was {spot.energy}, "
                    "but it does not exist. Aborting."
                )

            # set keys of phase space file to use
            source.position_key_x = self.position_key_x
            source.position_key_y = self.position_key_y
            source.position_key_z = self.position_key_z
            source.direction_key_x = self.direction_key_x
            source.direction_key_y = self.direction_key_y
            source.direction_key_z = self.direction_key_z
            source.energy_key = self.energy_key
            source.weight_key = self.weight_key
            source.PDGCode_key = self.PDGCode_key

            if self.batch_size is not None:
                source.batch_size = self.batch_size
            else:
                source.batch_size = 30000

            # if not set, initialize the entry_start to 0 or to a list for multithreading
            if self.entry_start is None:
                if not opengate_core.IsMultithreadedApplication():
                    self.entry_start = 0
                else:
                    # create a entry_start array with the correct number of start entries
                    # all entries are spaced by the number of particles/thread
                    n_threads = self.simulation.user_info.number_of_threads
                    # ui.entry_start = [0] * n_threads
                    step = np.ceil(nspot / n_threads) + 1  # Specify the increment value
                    self.entry_start = [i * step for i in range(n_threads)]
                print(
                    "INFO: entry_start not set. Using default values: ",
                    self.entry_start,
                )

            source.entry_start = self.entry_start

            # POSITION:
            source.translate_position = True
            source.position.translation = self._get_pbs_position(spot)
            # print("source.position.translation: ", source.position.translation)

            # ROTATION:
            source.rotate_direction = True
            # # use pbs rotation plus a potential rotation of the original phs
            # # may be necessary in case the original phs is not going into +z direction
            source.position.rotation = (
                self._get_pbs_rotation(spot) @ self.rotate_PhS_Source.as_matrix()
            )

            # add weight
            source.n = nspot

            # allow the possibility to count primaries
            source.generate_until_next_primary = self.generate_until_next_primary
            source.primary_lower_energy_threshold = self.primary_lower_energy_threshold
            source.primary_PDGCode = self.primary_PDGCode

        self.actual_sim_particles = tot_sim_particles

    def verify_phs_files_exist(self, phs_dict):
        """Check if all the files in the dictionary exist.
        Returns True if all the files exist, False otherwise
        If one file does not exist, the program is aborted."""

        for key, phs_file in phs_dict.items():
            # print(
            #     "key: ",
            #     key,
            #     " phs_file: ",
            #     phs_file,
            #     " type: ",
            #     type(phs_file),
            # )

            if not os.path.isfile(phs_file):
                fatal(
                    f"Error in ThreatmenPlanPhsSource: File {phs_file} does not exist. Aborting"
                )
        return True

    def read_list_of_Phs(self, file_name: str, path_to_phsp=""):
        """Read a list of Phs from a file.

        Parameters
        ----------
        file_name : str
            The name of the file to read from.
            File needs to have at least two columns
            First column is the energy in MeV which is the energy label
            Second column is the phase space file name
        Returns
        -------
        dictionary of Phs"""

        # prepend the path to the phase space files
        file_name = path_to_phsp + "/" + file_name

        phs_dict = {}
        try:
            input_arr = np.genfromtxt(
                file_name,
                delimiter="\t",
                comments="#",
                usecols=(0, 1),
                dtype=None,
                encoding=None,
            )

            # convert to dictionary
            if input_arr.shape == (0,):
                fatal(
                    f"Error in TreatmentPlanPhsSource: No data found in file: {file_name}. ",
                )
            if input_arr.ndim == 0:
                # only single line read, convert to array
                input_arr = np.array([input_arr])

            # normal case, multiple lines
            if path_to_phsp != "":
                phs_dict = {
                    float(i[0]): str(path_to_phsp + "/" + (i[1])) for i in input_arr
                }
            else:
                phs_dict = {float(i[0]): str(i[1]) for i in input_arr}
            # print("phs_dict read: ", phs_dict)
        except Exception as e:
            fatal(
                f"Error in TreatmentPlanPhsSource: Could not read the phase space file list. The error was: \n{e}"
            )
        if len(phs_dict) == 0:
            fatal(
                "Error in TreatmentPlanPhsSource: the phase space file list is empty. Aborting."
            )
        return phs_dict

    def verify_necessary_parameters_are_set(self):
        """This function checks whether all necessary parameters were set.
        E.g. not None
        It does not check sensibility"""
        if self.phaseSpaceList_file_name is None:
            fatal(
                "Error in TreatmentPlanPhsSource: phaseSpaceList_file_name is None. Aborting."
            )
        if self.distance_source_to_isocenter is None:
            fatal(
                "Error in TreatmentPlanPhsSource: distance_source_to_isocenter is None. Aborting."
            )
        if self.distance_stearmag_to_isocenter_x is None:
            fatal(
                "Error in TreatmentPlanPhsSource: distance_stearmag_to_isocenter_x is None. Aborting."
            )
        if self.distance_stearmag_to_isocenter_y is None:
            fatal(
                "Error in TreatmentPlanPhsSource: distance_stearmag_to_isocenter_y is None. Aborting."
            )
        if self.batch_size is None:
            fatal("Error in TreatmentPlanPhsSource: batch_size is None. Aborting.")
        if self.spots is None:
            fatal("Error in TreatmentPlanPhsSource: No spots have been set. Aborting.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.sources.utility import get_rad_yield

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test013_hl")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.check_volumes_overlap = False
    sim.output_dir = paths.output
    sim.random_seed = 42

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.s

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add two water boxes
    wb1 = sim.add_volume("Box", "waterbox1")
    wb1.size = [20 * cm, 20 * cm, 20 * cm]
    wb1.translation = [-20 * cm, 0, 0]
    wb2 = sim.add_volume("Box", "waterbox2")
    wb2.size = [20 * cm, 20 * cm, 20 * cm]
    wb2.translation = [20 * cm, 0, 0]

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = True
    sim.physics_manager.global_production_cuts.all = 0.1 * mm
    # p.energy_range_min = 250 * eV
    # p.energy_range_max = 15 * MeV

    # sources info
    activity = 10 * Bq
    hl = 6586.26 * sec  # 109.771 minutes

    # source ion
    ion_src = sim.add_source("GenericSource", "ion_source")
    ion_src.attached_to = wb1.name
    ion_src.particle = "ion 9 18"  # F18
    ion_src.position.type = "sphere"
    ion_src.position.radius = 10 * mm
    ion_src.direction.type = "iso"
    ion_src.energy.type = "mono"
    ion_src.energy.mono = 0
    ion_src.half_life = hl
    ion_src.activity = activity

    # source e+
    beta_src = sim.add_source("GenericSource", "beta+_source")
    beta_src.attached_to = wb2.name
    beta_src.particle = "e+"
    beta_src.position.type = "sphere"
    beta_src.position.radius = 10 * mm
    beta_src.energy.type = "F18"
    beta_src.direction.type = "iso"
    beta_src.half_life = hl
    total_yield = get_rad_yield("F18")
    print(f"{total_yield=}")
    beta_src.activity = activity * total_yield

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # filter for the phase space actor below
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "e+"

    # phsp
    phsp1 = sim.add_actor("PhaseSpaceActor", "phsp_ion")
    phsp1.attached_to = wb1.name
    phsp1.attributes = [
        "KineticEnergy",
        "LocalTime",
        "GlobalTime",
        "TrackProperTime",
        "TimeFromBeginOfEvent",
        # 'TrackVertexKineticEnergy', 'EventKineticEnergy'
    ]
    phsp1.output_filename = "test013_decay_ion.root"
    phsp1.steps_to_store = "first"
    phsp1.filters.append(f)

    phsp2 = sim.add_actor("PhaseSpaceActor", "phsp_beta")
    phsp2.attached_to = wb2.name
    phsp2.attributes = phsp1.attributes
    phsp2.output_filename = "test013_decay_beta_plus.root"
    phsp2.filters.append(f)
    phsp2.steps_to_store = "first"

    # long run
    sim.run_timing_intervals = [[0, 109 * 60 * sec]]

    # start simulation
    sim.run()

    # print results
    print(stats)

    print()
    keys1 = phsp1.attributes
    keys2 = keys1
    scalings = [1] * len(keys1)
    scalings[2] = 1e-12  # GlobalTime
    tols = [0.02] * len(keys1)
    # tols[1] = 0.02  # LocalTime
    tols[2] = 0.06  # GlobalTime
    # tols[4] = 0.02  # TimeFromBeginOfEvent
    print(keys2, scalings, tols)
    print(phsp1.get_output_path())
    print(phsp2.get_output_path())
    print()
    is_ok = utility.compare_root3(
        phsp1.get_output_path(),
        phsp2.get_output_path(),
        "phsp_ion",
        "phsp_beta",
        keys1,
        keys2,
        tols,
        scalings,
        scalings,
        paths.output / "test013_decay.png",
    )

    utility.test_ok(is_ok)



import matplotlib.pyplot as plt
import numpy as np
import matplotlib.patches as patches
import matplotlib.cm as cm
from box import Box
import opengate_core as g4
from opengate.geometry.utility import vec_g4_as_np, rot_g4_as_np
import json


def store_volumes_info(sim, json_filename):
    # get the pointers to the Geant4 volumes
    vol_info_raw = get_g4_volumes_pointers(sim)

    # Prepare a dictionary to hold processed volume data with world coordinates
    processed_volumes = serialize_volume_data(vol_info_raw)

    # save as json
    with open(json_filename, "w") as f:
        json.dump(processed_volumes, f, indent=4)


# Helper function to get the 8 corners of a bounding box
def _get_bbox_corners(xmin, ymin, zmin, xmax, ymax, zmax):
    """Returns the 8 corners of a bounding box."""
    corners = np.array(
        [
            [xmin, ymin, zmin, 1],
            [xmax, ymin, zmin, 1],
            [xmin, ymax, zmin, 1],
            [xmin, ymin, zmax, 1],
            [xmax, ymax, zmin, 1],
            [xmin, ymax, zmax, 1],
            [xmax, ymin, zmax, 1],
            [xmax, ymax, zmax, 1],
        ]
    )
    return corners


# Helper function to create a 4x4 homogeneous transformation matrix
def _create_transform_matrix(rotation_list, translation_list):
    """
    Creates a 4x4 homogeneous transformation matrix from a 3x3 rotation matrix
    and a 3-element translation vector.
    """
    transform_matrix = np.identity(4)
    if rotation_list is not None:
        transform_matrix[:3, :3] = np.array(rotation_list)
    if translation_list is not None:
        transform_matrix[:3, 3] = np.array(translation_list)
    return transform_matrix


# Recursive function to get the transformation matrix from a volume's local frame to the world frame
def _get_world_transform_matrix(volume_name, all_volume_data, memo={}):
    """
    Recursively calculates the cumulative transformation matrix from a volume's
    local coordinate system to the world coordinate system.
    Uses memoization to avoid redundant calculations.
    """
    if volume_name in memo:
        return memo[volume_name]

    if volume_name not in all_volume_data:
        # This case should ideally not happen if data is consistent, but handle defensively
        print(
            f"Warning: Volume '{volume_name}' not found in provided data. Assuming identity transform."
        )
        return np.identity(4)

    vol_info = all_volume_data[volume_name]

    # Get the transformation of the current volume relative to its mother
    local_rotation = vol_info.get("object_rotation")
    local_translation = vol_info.get("object_translation")

    # If object_translation/rotation is None, it means identity for placement
    if local_rotation is None:
        local_rotation = np.identity(3).tolist()
    if local_translation is None:
        local_translation = [0.0, 0.0, 0.0]

    current_transform = _create_transform_matrix(local_rotation, local_translation)

    mother_name = vol_info.get("mother")

    if mother_name is None:  # This is a top-level volume (like 'world' or 'arf_world')
        world_transform = current_transform
    else:
        # Recursively get the mother's transform to the world
        mother_world_transform = _get_world_transform_matrix(
            mother_name, all_volume_data, memo
        )
        # Combine current transform with mother's world transform
        world_transform = np.dot(mother_world_transform, current_transform)

    memo[volume_name] = world_transform
    return world_transform


def get_g4_volumes_pointers(sim):
    store = g4.G4PhysicalVolumeStore.GetInstance()
    map = store.GetMap()
    volumes = Box()
    for volume_name, volume_list in map.items():
        v = Box()
        v.physical_volume = volume_list[0]
        v.logical_volume = v.physical_volume.GetLogicalVolume()
        v.solid = v.logical_volume.GetSolid()
        volumes[str(volume_name)] = v

    return volumes


def serialize_volume_data(vol_info_raw):
    # Prepare a dictionary to hold processed volume data with world coordinates
    processed_volumes = Box()

    # To calculate world coordinates, serialize_volume_data needs access to
    # the object_translation and object_rotation of all volumes, including mothers.
    # We create a simplified dictionary for the recursive world transform calculation.
    simplified_all_data_for_transform = {
        name: {
            "mother": (
                str(data.physical_volume.GetMotherLogical().GetName())
                if data.physical_volume.GetMotherLogical()
                else None
            ),
            "object_translation": (
                vec_g4_as_np(data.physical_volume.GetObjectTranslation()).tolist()
                if data.physical_volume.GetObjectTranslation()
                else None
            ),
            "object_rotation": (
                rot_g4_as_np(data.physical_volume.GetObjectRotation()).tolist()
                if data.physical_volume.GetObjectRotation()
                else None
            ),
        }
        for name, data in vol_info_raw.items()
    }

    # Iterate through each volume and serialize its data, including world limits
    for k, v_raw in vol_info_raw.items():
        # Pass the raw volume info object and the simplified all_volume_data for recursion
        serialized_data = serialize_one_volume_data(
            v_raw, all_volume_data=simplified_all_data_for_transform
        )
        processed_volumes[k] = serialized_data

    return processed_volumes


def serialize_one_volume_data(vol_info, all_volume_data=None):
    """
    Serializes volume information, including local and world coordinate system
    bounding limits.
    """
    v = Box()
    # PL
    pl = vol_info.physical_volume
    v.pv_name = str(pl.GetName())
    v.is_replicated = pl.IsReplicated()
    v.is_parameterised = pl.IsParameterised()
    m = pl.GetMotherLogical()
    if m:
        v.mother = str(pl.GetMotherLogical().GetName())
    else:
        v.mother = None
    v.copy_no = pl.GetCopyNo()
    v.object_translation = None
    v.object_rotation = None
    v.frame_translation = None
    v.frame_rotation = None
    if pl.GetObjectTranslation() is not None:
        v.object_translation = vec_g4_as_np(pl.GetObjectTranslation()).tolist()
    if pl.GetObjectRotation() is not None:
        v.object_rotation = rot_g4_as_np(pl.GetObjectRotation()).tolist()
    if pl.GetFrameTranslation() is not None:
        v.frame_translation = vec_g4_as_np(pl.GetFrameTranslation()).tolist()
    if pl.GetFrameRotation() is not None:
        v.frame_rotation = rot_g4_as_np(pl.GetFrameRotation()).tolist()

    # LV
    lv = vol_info.logical_volume
    v.lv_name = str(lv.GetName())
    v.n_daughters = lv.GetNoDaughters()
    v.total_volume_entities = lv.TotalVolumeEntities()
    if lv.GetMaterial() is not None:
        v.material = str(lv.GetMaterial().GetName())
    else:
        v.material = None
    v.daughters = [str(lv.GetDaughter(i).GetName()) for i in range(lv.GetNoDaughters())]
    try:
        if v.material:
            v.mass = lv.GetMass(False, True, None)
    except:
        v.mass = None
    v.instance_id = lv.GetInstanceID()

    # Solid
    solid = vol_info.solid
    v.solid_name = str(solid.GetName())
    pMin_local = g4.G4ThreeVector()
    pMax_local = g4.G4ThreeVector()
    solid.BoundingLimits(pMin_local, pMax_local)

    # Store local bounding limits
    v.bounding_limits_local = [
        pMin_local.x,
        pMin_local.y,
        pMin_local.z,
        pMax_local.x,
        pMax_local.y,
        pMax_local.z,
    ]

    # Calculate world bounding limits if all_volume_data is provided
    v.world_bounding_limits = None
    if all_volume_data is not None:
        local_corners = _get_bbox_corners(*v.bounding_limits_local)
        world_transform_matrix = _get_world_transform_matrix(v.pv_name, all_volume_data)

        transformed_corners = np.dot(
            local_corners, world_transform_matrix.T
        )  # Apply transform

        # Extract min/max from transformed corners
        world_xmin, world_ymin, world_zmin = np.min(transformed_corners[:, :3], axis=0)
        world_xmax, world_ymax, world_zmax = np.max(transformed_corners[:, :3], axis=0)

        v.world_bounding_limits = [
            world_xmin,
            world_ymin,
            world_zmin,
            world_xmax,
            world_ymax,
            world_zmax,
        ]

    v.cubic_volume = solid.GetCubicVolume()
    v.surface_area = solid.GetSurfaceArea()
    v.entity_type = str(solid.GetEntityType())
    v.is_faceted = solid.IsFaceted()
    v.num_of_constituents = solid.GetNumOfConstituents()

    return v


def plot_volume_boundaries(
    volume_data: dict, axis: str = "X", show_labels: bool = True
):
    """
    Plots the bounding box boundaries of Geant4 volumes along a specified axis.
    """
    axis = axis.upper()
    if axis not in ["X", "Y", "Z"]:
        print(f"Error: Invalid axis '{axis}'. Please choose 'X', 'Y', or 'Z'.")
        return

    axis_map = {"X": (0, 3), "Y": (1, 4), "Z": (2, 5)}
    min_idx, max_idx = axis_map[axis]

    fig, ax = plt.subplots(figsize=(12, 8))

    plot_data = []
    for name, info in volume_data.items():
        if "bounding_limits" in info and len(info["bounding_limits"]) == 6:
            limits = info["bounding_limits"]
            min_val = limits[min_idx]
            max_val = limits[max_idx]
            plot_data.append({"name": name, "min": min_val, "max": max_val})
        else:
            print(
                f"Warning: Volume '{name}' missing 'bounding_limits' or invalid format. Skipping."
            )

    plot_data.sort(key=lambda x: x["min"])
    y_pos = np.arange(len(plot_data))

    for i, vol in enumerate(plot_data):
        ax.plot(
            [vol["min"], vol["max"]],
            [y_pos[i], y_pos[i]],
            marker="o",
            linestyle="-",
            markersize=8,
            label=vol["name"],
        )
        ax.text(
            vol["min"],
            y_pos[i],
            f'{vol["min"]:.2f}',
            va="center",
            ha="right",
            fontsize=8,
            color="blue",
        )
        ax.text(
            vol["max"],
            y_pos[i],
            f'{vol["max"]:.2f}',
            va="center",
            ha="left",
            fontsize=8,
            color="red",
        )
        if show_labels:
            ax.text(
                vol["max"] + (vol["max"] - vol["min"]) * 0.05,
                y_pos[i],
                vol["name"],
                va="center",
                fontsize=9,
                color="black",
            )

    ax.set_yticks(y_pos)
    ax.set_yticklabels([vol["name"] for vol in plot_data])
    ax.set_xlabel(f"Position along {axis}-axis")
    ax.set_ylabel("Volume Name")
    ax.set_title(f"Geant4 Volume Boundaries along {axis}-axis")
    ax.grid(True, linestyle="--", alpha=0.7)
    plt.tight_layout()
    plt.show()


def plot_volume_boundaries_2d(
    volume_data: dict,
    plane: str = "XY",
    show_labels: bool = True,
    highlight_volumes: list = None,
    ax=None,
):
    """
    Plots the 2D bounding boxes of Geant4 volumes on a specified plane.
    Can plot on a given matplotlib axis `ax` or create a new figure.
    """
    is_standalone = ax is None
    if is_standalone:
        fig, ax = plt.subplots(figsize=(12, 10))

    plane = plane.upper()
    if plane not in ["XY", "XZ", "YZ"]:
        print(f"Error: Invalid plane '{plane}'. Please choose 'XY', 'XZ', or 'YZ'.")
        return

    highlight_volumes = highlight_volumes or []

    plane_map = {
        "XY": ((0, 1), (3, 4), "X-axis", "Y-axis"),
        "XZ": ((0, 2), (3, 5), "X-axis", "Z-axis"),
        "YZ": ((1, 2), (4, 5), "Y-axis", "Z-axis"),
    }
    (min1_idx, min2_idx), (max1_idx, max2_idx), xlabel, ylabel = plane_map[plane]

    def _get_volume_depth(volume_name, data, current_depth=0, memo={}):
        if volume_name in memo:
            return memo[volume_name]
        if volume_name not in data or data[volume_name].get("mother") is None:
            memo[volume_name] = current_depth
            return current_depth
        else:
            mother_name = data[volume_name]["mother"]
            depth = _get_volume_depth(mother_name, data, current_depth + 1, memo)
            memo[volume_name] = depth
            return depth

    processed_volumes = []
    for name, info in volume_data.items():
        if (
            "world_bounding_limits" in info
            and info["world_bounding_limits"]
            and len(info["world_bounding_limits"]) == 6
        ):
            depth = _get_volume_depth(name, volume_data)
            processed_volumes.append({"name": name, "info": info, "depth": depth})
        else:
            print(
                f"Warning: Volume '{name}' missing 'world_bounding_limits' or invalid format. Skipping for 2D plot."
            )

    processed_volumes.sort(key=lambda x: (x["depth"], x["name"]))

    volumes_to_plot_normal = [
        v for v in processed_volumes if v["name"] not in highlight_volumes
    ]
    volumes_to_plot_highlight = [
        v for v in processed_volumes if v["name"] in highlight_volumes
    ]
    plotting_order = volumes_to_plot_normal + volumes_to_plot_highlight

    num_volumes_to_plot = len(plotting_order)
    colors_cmap = cm.get_cmap(
        "tab20" if num_volumes_to_plot <= 20 else "viridis", num_volumes_to_plot
    )
    color_idx = 0
    all_x_coords, all_y_coords = [], []

    for vol_entry in plotting_order:
        name, info = vol_entry["name"], vol_entry["info"]
        limits = info["world_bounding_limits"]
        x_min, y_min = limits[min1_idx], limits[min2_idx]
        x_max, y_max = limits[max1_idx], limits[max2_idx]
        width, height = x_max - x_min, y_max - y_min
        color = colors_cmap(color_idx)
        color_idx += 1
        linewidth, edgecolor = (
            (3.0, "black") if name in highlight_volumes else (1.5, color)
        )

        rect = patches.Rectangle(
            (x_min, y_min),
            width,
            height,
            linewidth=linewidth,
            edgecolor=edgecolor,
            facecolor="none",
            label=name,
        )
        ax.add_patch(rect)

        # Add text labels for min and max values on the axes
        if width > 0 and height > 0:
            ax.text(
                x_min,
                y_min - height * 0.03,
                f"{x_min:.1f}",
                ha="center",
                va="top",
                fontsize=7,
                color=edgecolor,
            )
            ax.text(
                x_max,
                y_min - height * 0.03,
                f"{x_max:.1f}",
                ha="center",
                va="top",
                fontsize=7,
                color=edgecolor,
            )
            ax.text(
                x_min - width * 0.03,
                y_min,
                f"{y_min:.1f}",
                ha="right",
                va="center",
                fontsize=7,
                color=edgecolor,
            )
            ax.text(
                x_min - width * 0.03,
                y_max,
                f"{y_max:.1f}",
                ha="right",
                va="center",
                fontsize=7,
                color=edgecolor,
            )

        if show_labels:
            ax.text(
                x_max + width * 0.02,
                y_min + height / 2,
                name,
                ha="left",
                va="center",
                fontsize=8,
                color=color,
                bbox=dict(
                    facecolor="white",
                    alpha=0.7,
                    edgecolor="none",
                    boxstyle="round,pad=0.2",
                ),
            )

        if "world" not in name.lower():
            all_x_coords.extend([x_min, x_max])
            all_y_coords.extend([y_min, y_max])

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(f"Volume Boundaries in {plane}-plane")
    ax.grid(True, linestyle="--", alpha=0.7)
    ax.set_aspect("equal", adjustable="box")

    if all_x_coords and all_y_coords:
        x_min_plot, x_max_plot = min(all_x_coords), max(all_x_coords)
        y_min_plot, y_max_plot = min(all_y_coords), max(all_y_coords)
        x_buffer = (x_max_plot - x_min_plot) * 0.1
        y_buffer = (y_max_plot - y_min_plot) * 0.1
        ax.set_xlim(x_min_plot - x_buffer, x_max_plot + x_buffer)
        ax.set_ylim(y_min_plot - y_buffer, y_max_plot + y_buffer)
    else:
        ax.autoscale_view()

    if is_standalone:
        handles, labels = ax.get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        if by_label:
            ax.legend(
                by_label.values(),
                by_label.keys(),
                loc="center left",
                bbox_to_anchor=(1.02, 0.5),
                fontsize=8,
            )
        # plt.tight_layout(rect=[0, 0, 0.85, 1])
        plt.show()


def plot_all_views_2d(
    volume_data: dict, show_labels: bool = True, highlight_volumes: list = None
):
    """
    Plots the 2D bounding boxes of Geant4 volumes from three different views (XY, XZ, YZ)
    on a single figure with three subplots.
    """
    fig, axes = plt.subplots(1, 3, figsize=(30, 9))

    # Plot on each subplot
    plot_volume_boundaries_2d(
        volume_data,
        plane="XY",
        show_labels=show_labels,
        highlight_volumes=highlight_volumes,
        ax=axes[0],
    )
    plot_volume_boundaries_2d(
        volume_data,
        plane="XZ",
        show_labels=show_labels,
        highlight_volumes=highlight_volumes,
        ax=axes[1],
    )
    plot_volume_boundaries_2d(
        volume_data,
        plane="YZ",
        show_labels=show_labels,
        highlight_volumes=highlight_volumes,
        ax=axes[2],
    )

    # Create a single, combined legend for the entire figure
    handles, labels = axes[0].get_legend_handles_labels()
    by_label = dict(zip(labels, handles))  # To remove duplicate labels
    if by_label:
        fig.legend(
            by_label.values(),
            by_label.keys(),
            loc="center right",
            bbox_to_anchor=(0.98, 0.5),
            fontsize=9,
        )

    fig.suptitle("Geant4 Volume Boundaries - All Views", fontsize=16)
    # plt.tight_layout(rect=[0, 0, 0.9, 0.95])  # Adjust for suptitle and legend
    plt.show()



import tables
from anytree import NodeMixin
from pathlib import Path

from .utility import (
    get_node_name,
    get_parent,
    get_node_path,
    get_create_function,
    set_node_name,
)
from ..base import GateObject
from ..exception import fatal
from ..utility import ensure_directory_exists

__run_group_prefix__ = "data_from_run"


def _setter_hook_yields_to(self, yields_to):
    try:
        unit_name = yields_to.name
    except AttributeError:
        unit_name = yields_to
    if self.processing_group is not None:
        self.processing_group._need_tree_update = True
    return unit_name


class ProcessingUnitBase(GateObject, NodeMixin):
    user_info_defaults = {
        "yields_to": (
            None,
            {
                "doc": "Processing unit that provides the input to this processing unit. ",
                "setter_hook": _setter_hook_yields_to,
            },
        ),
        "input_name": (
            None,
            {
                "doc": "Identifiers to specify the input data, usually node names. ",
            },
        ),
    }

    def __init__(self, *args, processing_group=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.processing_group = processing_group
        self.output_data_handles = {}
        self.input_data_handles = {}
        self.hdf5_group = None  # the hdf5 group where this unit stores its data

    @property
    def post_processor(self):
        if self.processing_group.is_post_processor:
            return self.processing_group
        else:
            return self.processing_group.post_processor

    @property
    def is_processing_unit(self):
        return True

    @property
    def can_be_initial_unit(self):
        """Generally, processing units cannot be the initial units in a processing tree.
        Those units which can be initial units, e.g. data fetcher, should override this property.
        """
        return False

    @property
    def output_file_handle(self):
        return self.post_processor.output_file_handle

    @property
    def input_units(self):
        if len(self.children) > 0:
            return self.children
        else:
            if self.processing_group is not None:
                try:
                    return self.processing_group.input_units
                except AttributeError:
                    return tuple()
            else:
                return tuple()

    def get_or_create_output_table(self, *args, **kwargs):
        return self._get_or_create_output_structure("tables", *args, **kwargs)

    def get_or_create_output_array(self, *args, **kwargs):
        return self._get_or_create_output_structure("array", *args, **kwargs)

    def get_or_create_output_carray(self, *args, **kwargs):
        return self._get_or_create_output_structure("carray", *args, **kwargs)

    def get_or_create_output_earray(self, *args, **kwargs):
        return self._get_or_create_output_structure("earray", *args, **kwargs)

    def get_or_create_output_group(self, *args, **kwargs):
        return self._get_or_create_output_structure("group", *args, **kwargs)

    def _get_or_create_output_structure(
        self, which, name, link_name=None, subgroup=None, external_file=False, **kwargs
    ):
        g = self.get_unit_output_group()
        if subgroup is not None:
            try:
                g = self.output_file_handle.get_node(g, subgroup)
            except tables.NoSuchNodeError:
                try:
                    g = self.output_file_handle.create_group(
                        g, subgroup, createparents=True
                    )
                except tables.NodeError as e:
                    fatal(
                        f"Could not create subgroup while creating table. The following exception occurred: "
                        f"{str(e)}"
                    )
        if external_file is False:
            try:
                return self.output_file_handle.get_node(g, name)
            except tables.NoSuchNodeError:
                return get_create_function(self.output_file_handle, which)(
                    g, name, **kwargs
                )
        else:
            if link_name is None:
                fatal(
                    f"You must provide a link_name when the option external_file=True. "
                )
            try:
                return self.output_file_handle.get_node(g, link_name)
            except tables.NoSuchNodeError:
                output_folder = (
                    self.post_processor.output_directory_external_files
                    / Path(get_node_path(g, strip_leading_slash=True))
                )
                ensure_directory_exists(output_folder)
                output_filename = f"{link_name}.h5"
                output_path = output_folder / output_filename
                print("output_path: ", output_path)
                f = self.post_processor.get_or_register_extra_file_handle(
                    tables.open_file(output_path, "w")
                )
                try:
                    s_external = f.get_node("/", name)
                except tables.NoSuchNodeError:
                    s_external = get_create_function(f, which)("/", name, **kwargs)
                # hdf5_path = get_node_path(s_external)
                link = self.post_processor.create_external_link(
                    g, link_name, s_external
                )  # this is a link object
                # g, name, f"{str(output_path)}:{hdf5_path}")  # this is a link object
                # node = link()  # l() returns the resolved link
                return link

    def register_output_data_handle(self, output_data_handle):
        """output_data_handle should be a mode in the hdf5_group belonging to this unit."""
        identifier = self.get_unique_identifier_within_unit(output_data_handle)
        # external links should need to be resolved before storage
        if hasattr(output_data_handle, "extfile"):
            _handle = output_data_handle(mode="a")
        else:
            _handle = output_data_handle
        if identifier not in self.output_data_handles:
            # set_node_name(_handle, identifier)  # don't change the target name!
            self.output_data_handles[identifier] = _handle
        else:
            fatal(f"An output data handle with this name already exists. ")
        return _handle

    def get_unique_identifier_within_unit(self, node):
        # construct a unique identifier.
        # If the output data handle is directly inside the output group of this unit,
        # the node name is guaranteed to be unique.
        # Otherwise, the subgroups needed to be considered as well (e.g. channels)
        items_to_combine = [node]
        counter = 0
        while get_parent(items_to_combine[-1]) is not self.hdf5_group:
            print(items_to_combine[-1])
            items_to_combine.append(get_parent(items_to_combine[-1]))
            if counter > 100:
                fatal(
                    f"Maximum recursion depth reached. "
                    f"Something went wrong while trying to find a unique identifier "
                    f"for node {get_node_name(items_to_combine[0])}"
                )
            counter += 1
        return "_".join([get_node_name(n) for n in items_to_combine])

    def get_title_for_hdf5_node(self):
        return f"{type(self).__name__}_{self.name}"

    def update_tree_node(self):
        """Internal method which retrieves the volume object
        from the volume manager based on the mother's name stored as user info 'mother'
        """
        if self.processing_group is not None:
            if self.yields_to is not None:
                try:
                    self.parent = self.processing_group.get_processing_unit(
                        self.yields_to
                    )
                except KeyError:
                    fatal(
                        f"Error while trying to update the hierarchy relationship of processing unit {self.name}. "
                        f"Input unit {self.yields_to} not found in processing group {self.processing_group.name}."
                    )
            else:
                self.parent = None
        else:
            fatal(
                f"Cannot update hierarchy of processing unit {self.name} "
                f"because it is not attached to any processing group. "
            )

    def _request_processing_unit_tree_update(self):
        if self.processing_group is not None:
            self.processing_group.update_processing_tree()
        else:
            fatal(
                f"Cannot update hierarchy of processing unit {self.name} "
                f"because it is not attached to any processing group. "
            )

    def initialize_hdf5_group_in_output_file(self):
        if self.hdf5_group is None:
            if self.processing_group.hdf5_group is None:
                self.processing_group.initialize_hdf5_group_in_output_file()
            self.hdf5_group = self.output_file_handle.create_group(
                self.processing_group.hdf5_group, self.name
            )

    def get_unit_output_group(self):
        return self.hdf5_group

    def store_user_attributes(self):
        attrs = self.get_unit_output_group()._v_attrs
        for k, v in self.to_dictionary().items():
            setattr(attrs, k, v)

    def close(self):
        super().close()
        for k in self.input_data_handles.keys():
            self.input_data_handles[k] = None
        for child in self.children:
            child.close()

    def update_output(self):
        # do not override this method in derived classes
        for child in self.children:
            child.update_output()
        self.do_your_job()
        self.store_user_attributes()
        print(f"updated {self.name}")

    def do_your_job(self):
        """Implement this in the concrete class."""
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate_core as g4
from opengate.tests import utility

if __name__ == "__main__":
    # get the attributes manager
    am = g4.GateDigiAttributeManager.GetInstance()

    # retrieve the list of attributes names
    nlist = am.GetAvailableDigiAttributeNames()

    # print
    print(f"List of all available attributes ({len(nlist)})")
    print(
        f"Types are: 3 (ThreeVector), D (double), S (string), I (int), U (unique volume ID)"
    )
    for a in nlist:
        att = am.GetDigiAttributeByName(a)
        print(att.GetDigiAttributeName(), att.GetDigiAttributeType())

    n = 52
    is_ok = len(nlist) == n

    utility.print_test(is_ok, f"Done for {n} attributes.")

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# config 1 is used for tests of "keepAll", "keepHighestEnergyPair", "removeAll" and "keepCoincidenceIfOnlyOneGood" policies

import opengate as gate
from pathlib import Path
import os
from opengate.tests import utility

# colors (similar to the ones of Gate)
red = [1, 0, 0, 1]
blue = [0, 0, 1, 1]
green = [0, 1, 0, 1]
yellow = [0.9, 0.9, 0.3, 1]
gray = [0.5, 0.5, 0.5, 1]
white = [1, 1, 1, 0.8]

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test072", "test072")

    sim = gate.Simulation()

    # options
    # warning the visualisation is slow !
    sim.visu = False
    sim.visu_type = "vrml"
    sim.random_seed = "auto"
    sim.number_of_threads = 1

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    sec = gate.g4_units.s
    ps = gate.g4_units.ps
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    gcm3 = gate.g4_units.g_cm3
    deg = gate.g4_units.deg

    # folders
    data_path = paths.data
    output_path = paths.output

    # world
    world = sim.world
    world.size = [450 * mm, 450 * mm, 70 * mm]
    world.material = "G4_AIR"

    # add the Philips Vereos PET
    # pet = pet_vereos.add_pet(sim, "pet")
    # if create_mat:
    #    create_material(sim)

    # create the material lead
    sim.volume_manager.material_database.add_material_nb_atoms(
        "Lead", ["Pb"], [1], 11.4 * gcm3
    )
    sim.volume_manager.material_database.add_material_nb_atoms(
        "Uranium", ["U"], [1], 18.9 * gcm3
    )
    sim.volume_manager.material_database.add_material_nb_atoms(
        "BGO", ["Bi", "Ge", "O"], [4, 3, 12], 7.13 * gcm3
    )

    sim.volume_manager.material_database.add_material_nb_atoms(
        "LSO", ["Lu", "Si", "O"], [2, 1, 5], 7.4 * gcm3
    )

    sim.volume_manager.material_database.add_material_weights(
        "LeadSb", ["Pb", "Sb"], [0.95, 0.05], 11.16 * gcm3
    )

    sim.volume_manager.material_database.add_material_weights(
        "LYSO",
        ["Lu", "Y", "Si", "O"],
        [0.31101534, 0.368765605, 0.083209699, 0.237009356],
        5.37 * gcm3,
    )

    # ring volume
    pet = sim.add_volume("Tubs", "pet")
    pet.rmax = 200 * mm
    pet.rmin = 127 * mm
    pet.dz = 32 * mm  # 340 * mm / 2.0
    pet.color = gray
    pet.material = "G4_AIR"

    # block
    block = sim.add_volume("Box", "block")
    block.mother = pet.name
    block.size = [60 * mm, 10 * mm, 10 * mm]
    # block.size = [1 * mm, 10 * mm, 10  * mm]
    # block.translation = [0 * mm, 324.3 * mm , 0 * mm]
    translations_ring, rotations_ring = gate.geometry.utility.get_circular_repetition(
        80, [160 * mm, 0.0 * mm, 0], start_angle_deg=180, axis=[0, 0, 1]
    )
    block.translation = translations_ring
    block.rotation = rotations_ring

    block.material = "G4_AIR"
    # block.translation = get_grid_repetition([1, 1, 4], [0, 0 * mm, 38.8 * mm])
    block.color = white

    # Crystal
    crystal = sim.add_volume("Box", "crystal")
    crystal.mother = block.name
    # crystal.size = [1 * mm, 10 * mm, 10 * mm]
    # crystal.material = "Uranium"
    crystal.size = [60 * mm, 10 * mm, 10 * mm]
    crystal.material = "LYSO"
    # crystal.translation = gate.geometry.utility.get_grid_repetition([1, 1, 1], [3.9833 * mm ,0 * mm, 5.3 * mm])
    crystal.color = green

    """
    # If visu is enabled, we simplified the PET system, otherwise it is too slow
    if sim.visu:
        module = sim.volume_manager.get_volume("module")
        # only 2 repetition instead of 18
        translations_ring, rotations_ring = get_circular_repetition(
        72, [427 * mm, 0, 0], start_angle_deg=190, axis=[0, 0, 1]
        )
        module.translation = translations_ring
        module.rotation = rotations_ring
    """

    source = sim.add_source("GenericSource", "b2b")
    source.particle = "back_to_back"
    source.activity = 200000000 * Bq  # 2000000 * Bq
    # source.n = 1000
    source.position.type = "sphere"
    source.position.radius = 0.0000000000000005 * mm
    # source.direction.type = "iso"
    # source.direction.accolinearity_flag = False
    source.energy.mono = 511 * keV
    source.direction.theta = [90 * deg, 90 * deg]
    source.direction.phi = [0, 360 * deg]

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    # sim.physics_manager.enable_decay = True
    # sim.physics_manager.set_production_cut("world", "all", 1 * m)
    # sim.physics_manager.set_production_cut("waterbox", "all", 1 * mm)

    # add the PET digitizer

    # Hits
    hc = sim.add_actor("DigitizerHitsCollectionActor", f"Hits_{crystal.name}")
    hc.attached_to = crystal.name
    hc.authorize_repeated_volumes = True
    hc.output_filename = output_path / "output_singles.root"
    # hc.output = "output_config1.root"
    hc.attributes = [
        "EventID",
        "PostPosition",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # Singles
    sc = sim.add_actor("DigitizerAdderActor", f"Singles_{crystal.name}")
    sc.attached_to = hc.attached_to
    sc.authorize_repeated_volumes = True
    sc.input_digi_collection = hc.name
    sc.policy = "EnergyWinnerPosition"
    sc.output_filename = hc.output_filename

    # timing
    sim.run_timing_intervals = [[0, 0.0002 * sec]]
    # sim.run_timing_intervals = [[0, 0.01 * sec]]
    # go
    sim.run()

    # end
    """print(f"Output statistics are in {stats.output}")
    print(f"Output edep map is in {dose.output}")
    print(f"vv {ct.image} --fusion {dose.output}")
    stats = sim.output.get_actor("Stats")
    print(stats)"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import itk
from scipy.spatial.transform import Rotation

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test021")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = 1
    sim.random_seed = 123456
    sim.output_dir = paths.output
    print(sim)

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    #  change world size
    sim.world.size = [1.5 * m, 1 * m, 1 * m]

    # fake box #1
    fake = sim.add_volume("Box", "fake")
    fake.size = [36 * cm, 36 * cm, 36 * cm]
    fake.translation = [25 * cm, 0, 0]
    r = Rotation.from_euler("y", -25, degrees=True)
    r = r * Rotation.from_euler("x", -35, degrees=True)
    fake.rotation = r.as_matrix()

    # ---------------------------------------------------
    # CT image #1
    ct = sim.add_volume("Image", "ct")
    ct.image = str(paths.data / "10x10x10.mhd")
    ct.mother = fake.name
    ct.voxel_materials = [[0, 10, "G4_WATER"]]
    ct.translation = [-3 * cm, 0, 0]
    r = Rotation.from_euler("z", 45, degrees=True)
    ct.rotation = r.as_matrix()

    ct_info = gate.image.read_image_info(ct.image)
    print(f"CT image origin and size: ", ct_info.origin, ct_info.size, ct_info.spacing)

    # source from image for CT #1
    source = sim.add_source("VoxelSource", "vox_source")
    source.attached_to = ct.name
    source.particle = "alpha"
    source.activity = 10000 * Bq / sim.number_of_threads
    source.image = str(paths.data / "five_pixels_10.mhd")
    source.direction.type = "iso"
    source.position.translation = gate.image.get_translation_between_images_center(
        ct.image, source.image
    )
    print(f"Source wrt CT 10x10x10 translation", source.position.translation)
    source.energy.mono = 1 * MeV
    src_info = gate.image.read_image_info(source.image)
    print(
        f"Source image origin and size: ",
        src_info.origin,
        src_info.size,
        src_info.spacing,
    )

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.edep.output_filename = "test021-1.mhd"
    # we do not need to write the image to disk
    # because we can get it directly from the actor after the simulation
    dose.edep.write_to_disk = False
    dose.attached_to = ct.name
    img_info = gate.image.read_image_info(ct.image)
    dose.size = img_info.size
    dose.spacing = img_info.spacing
    dose.output_coordinate_system = "attached_to_image"

    # cuts
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.all = 1 * mm
    # sim.physics_manager.set_production_cut("world", "all", 1 * mm)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # test pixels in dose #1
    # test pixels in dose #1
    d_even = dose.edep.get_data()
    s = itk.array_view_from_image(d_even).sum()
    v0 = d_even.GetPixel([5, 5, 5])
    v1 = d_even.GetPixel([1, 5, 5])
    v2 = d_even.GetPixel([1, 2, 5])
    v3 = d_even.GetPixel([5, 2, 5])
    v4 = d_even.GetPixel([6, 2, 5])
    tol = 0.15
    ss = v0 + v1 + v2 + v3 + v4

    def t(s, v):
        diff = abs(s - v) / s
        b = diff < tol
        p = diff * 100.0
        utility.print_test(b, f"Image diff {s:.2f} vs {v:.2f}  -> {p:.2f}%")
        return b

    is_ok = t(s, ss)
    is_ok = t(2000, v0) and is_ok
    is_ok = t(2000, v1) and is_ok
    is_ok = t(2000, v2) and is_ok
    is_ok = t(2000, v3) and is_ok
    is_ok = t(2000, v4) and is_ok

    stats_ref = utility.read_stat_file(paths.output_ref / "stat021_ref_1.txt")
    stats_ref.counts.runs = sim.number_of_threads
    is_ok = utility.assert_stats(stats, stats_ref, 0.1) and is_ok

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Context: See test079_acollin_ions_geant4_material.py

Here, the material, IEC_PLASTIC is not known from Geant4 and not defined in
GateMaterials.db, so one only need to add its definition to the volume_manager and then
set the mean energy per ion in the physics_manager
"""

from test079_acollin_helpers import *
import opengate.tests.utility as tu
import matplotlib.pyplot as plt


#########################################################################################
# Simulations configuration that may be relevant to change
#########################################################################################
# Mean energy of Ion Pair to use. 5.0 eV should produce the expected 0.5 deg FWHM in PET
# imaging
mean_energy = 5.0 * eV
# Key added to output to make sure that multi-threading the tests does not backfire
test_key = "p1"


#########################################################################################
# Main : We use this to launch the test
#########################################################################################
if __name__ == "__main__":
    paths = tu.get_default_test_paths(__file__, output_folder="test079")

    # Define core of the simulation, including physics
    sim = setup_simulation_engine(paths)

    # add a waterbox
    wb = sim.add_volume("Box", "waterbox")
    wb.size = [50 * cm, 50 * cm, 50 * cm]
    elems = ["C", "H", "O"]
    nbAtoms = [5, 8, 2]
    sim.volume_manager.material_database.add_material_nb_atoms(
        "IEC_PLASTIC", elems, nbAtoms, 1.18 * gcm3
    )
    wb.material = "IEC_PLASTIC"

    # set the source
    source = sim.add_source("GenericSource", "f18")
    source.particle = "ion 9 18"
    source.energy.mono = 0
    source.activity = 10000 * Bq
    source.direction.type = "iso"

    # add phase actor
    phsp = setup_actor(sim, "phsp", wb.name)
    phsp.output_filename = paths.output / f"annihilation_photons_{test_key}.root"

    # go
    sim.run(start_new_process=True)

    # redo test changing the MeanEnergyPerIonPair
    root_filename = phsp.output_filename
    phsp.output_filename = (
        paths.output / f"annihilation_photons_with_mepip_{test_key}.root"
    )

    sim.physics_manager.mean_energy_per_ion_pair["IEC_PLASTIC"] = mean_energy

    # go
    sim.run()

    # test: no mean energy, should be mostly colinear
    gamma_pairs = read_gamma_pairs(root_filename)
    acollinearity_angles = compute_acollinearity_angles(gamma_pairs)

    colin_median = plot_colin_case(acollinearity_angles)

    # test: with mean energy, acolinearity amplitude should have a Rayleigh distribution
    gamma_pairs = read_gamma_pairs(phsp.output_filename)
    acollinearity_angles = compute_acollinearity_angles(gamma_pairs)

    acolin_scale = plot_acolin_case_mepip(mean_energy, acollinearity_angles)

    f = paths.output / f"acollinearity_angles_{test_key}.png"
    plt.savefig(f)
    print(f"Plot was saved in {f}")

    # final
    # No acolin
    is_ok_p1 = colin_median < 0.01
    # Basic acolin
    is_ok_p2 = np.isclose(acolin_scale * 2.355, 0.5, atol=0.2)

    tu.test_ok(is_ok_p1 and is_ok_p2)



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 11:52:20 2015
Function providing the mass attenuation coefficient, the mass energy absorption coefficient
and the linear attenuation coefficient of a mixture for energies in [0.001,20] MeV range.
@author: vilches
"""

from .PhotonAttenuation import *


def PhotonAttenuationMixture(Mixture, E, Option):
    [w, El] = ChComposition(
        Mixture
    )  # extract fraction by weight and elements of the mixture
    d = PropsMix[
        PropsMix[:, 2] == Mixture, 1
    ]  # extract the density of the mixture from PropsMix
    Z = [0 for i in range(len(El))]
    for i in range(0, len(El)):
        Z[i] = np.where(PropsEl[:, 2] == El[i])
    ix = [x[0] for x in Z]
    idx = list(map(int, ix))  # assign atomic number value to each element
    w = list(map(float, np.asarray(w)))
    d = list(map(float, d))
    murho = [0 for j in range(len(idx))]
    murhoMix = [0 for j in range(len(idx))]
    muMix = [0 for j in range(len(idx))]
    muMixture = [0 for k in range(len(idx))]

    if Option == 1:  # mass attenuation coefficient of the mixture
        for i in range(0, len(idx)):
            murho[i] = PhotonAttenuationEl((idx[i] + 1), E, 1)  # mac
            murhoMix[i] = np.dot(murho[i], w[i])
        muMix = np.multiply(np.asarray(murhoMix), 1)
        muMixture = muMix.sum(axis=0)

    elif Option == 2:  # mass energy absorption coefficient of the mixture
        for i in range(0, len(idx)):
            murho[i] = PhotonAttenuationEl((idx[i] + 1), E, 2)  # meac
            murhoMix[i] = np.dot(murho[i], w[i])
        muMix = np.multiply(np.asarray(murhoMix), 1)
        muMixture = muMix.sum(axis=0)

    elif Option == 3:  # linear attenuation coefficient of the mixture
        for i in range(0, len(idx)):
            murho[i] = PhotonAttenuationEl((idx[i] + 1), E, 1)  # mac
            murhoMix[i] = np.dot(murho[i], w[i])
        muMix = np.multiply(np.asarray(murhoMix), d)  # from mac to lac
        muMixture = muMix.sum(axis=0)
    elif Option > 3:
        print("Error! Only three outputs are possible!")
        return ["Set the function inputs again"]
    return muMixture


# Example 1:
# E = 0.001
# result = PhotonAttenuationMixture('Soft Tissue', E, 3)
# print result

## Example 2:
# E = np.arange(0.001,0.01,0.001) # CT imaging range: from 1 keV to 150 keV
# result = [ 0 for i in range(len(E))]
# for i in range(0,len(E)):
#    result = PhotonAttenuationMixture('Air', E, 2)
# print result



from box import Box
import platform
import opengate_core as g4
from .base import ActorBase
from ..utility import g4_units, g4_best_unit_tuple
from .actoroutput import ActorOutputBase, ActorOutputSingleImage
from ..serialization import dump_json
from ..exception import fatal, warning
from ..base import process_cls

"""
    It is feasible to get callback every Run, Event, Track, Step in the python side.
    However, it is VERY time consuming. For SteppingAction, expect large performance drop.
    It could be however useful for prototyping or tests.

    it requires "trampoline functions" on the cpp side.

    # it is feasible but very slow !
    def SteppingAction(self, step, touchable):
        g4.GateSimulationStatisticsActor.SteppingAction(self, step, touchable)
        do_something()
"""


def _setter_hook_stats_actor_output_filename(self, output_filename):
    # By default, write_to_disk is False.
    # However, if user actively sets the output_filename
    # s/he most likely wants to write to disk also
    if output_filename != "" and output_filename is not None:
        self.write_to_disk = True
    return output_filename


class ActorOutputStatisticsActor(ActorOutputBase):
    """This is a hand-crafted ActorOutput specifically for the SimulationStatisticsActor."""

    # hints for IDE
    encoder: str
    output_filename: str
    write_to_disk: bool

    user_info_defaults = {
        "encoder": (
            "json",
            {
                "doc": "How should the output be encoded?",
                "allowed_values": ("json", "legacy"),
            },
        ),
        "output_filename": (
            "auto",
            {
                "doc": "Filename for the data represented by this actor output. "
                "Relative paths and filenames are taken "
                "relative to the global simulation output folder "
                "set via the Simulation.output_dir option. ",
                "setter_hook": _setter_hook_stats_actor_output_filename,
            },
        ),
        "write_to_disk": (
            False,
            {
                "doc": "Should the output be written to disk, or only kept in memory? ",
            },
        ),
        "active": (
            True,
            {"doc": "This actor is always active. ", "read_only": True},
        ),
    }

    default_suffix = "json"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # predefine the merged_data
        self.merged_data = Box()
        self.merged_data.runs = 0
        self.merged_data.events = 0
        self.merged_data.tracks = 0
        self.merged_data.steps = 0
        self.merged_data.duration = 0
        self.merged_data.start_time = 0
        self.merged_data.stop_time = 0
        self.merged_data.sim_start_time = 0
        self.merged_data.sim_stop_time = 0
        self.merged_data.init = 0
        self.merged_data.track_types = {}
        self.merged_data.nb_threads = 1

    @property
    def pps(self):
        if self.merged_data.duration != 0:
            return int(
                self.merged_data.events / (self.merged_data.duration / g4_units.s)
            )
        else:
            return 0

    @property
    def tps(self):
        if self.merged_data.duration != 0:
            return int(
                self.merged_data.tracks / (self.merged_data.duration / g4_units.s)
            )
        else:
            return 0

    @property
    def sps(self):
        if self.merged_data.duration != 0:
            return int(
                self.merged_data.steps / (self.merged_data.duration / g4_units.s)
            )
        else:
            return 0

    def store_data(self, data, **kwargs):
        self.merged_data.update(data)

    def get_data(self, **kwargs):
        if "which" in kwargs and kwargs["which"] != "merged":
            warning(
                f"The statistics actor output only stores merged data currently. "
                f"The which={kwargs['which']} you provided will be ignored. "
            )
        # the statistics actor currently only handles merged data, so we return it
        # no input variable 'which' as in other output classes
        return self.merged_data

    def get_processed_output(self):
        d = {}
        d["runs"] = {"value": self.merged_data.runs, "unit": None}
        d["events"] = {"value": self.merged_data.events, "unit": None}
        d["tracks"] = {"value": self.merged_data.tracks, "unit": None}
        d["steps"] = {"value": self.merged_data.steps, "unit": None}
        val, unit = g4_best_unit_tuple(self.merged_data.init, "Time")
        d["init"] = {
            "value": val,
            "unit": unit,
        }
        val, unit = g4_best_unit_tuple(self.merged_data.duration, "Time")
        d["duration"] = {
            "value": val,
            "unit": unit,
        }
        d["pps"] = {"value": self.pps, "unit": None}
        d["tps"] = {"value": self.tps, "unit": None}
        d["sps"] = {"value": self.sps, "unit": None}
        d["start_time"] = {
            "value": self.merged_data.start_time,
            "unit": None,
        }
        d["stop_time"] = {
            "value": self.merged_data.stop_time,
            "unit": None,
        }
        val, unit = g4_best_unit_tuple(self.merged_data.sim_start_time, "Time")
        d["sim_start_time"] = {
            "value": val,
            "unit": unit,
        }
        val, unit = g4_best_unit_tuple(self.merged_data.sim_stop_time, "Time")
        d["sim_stop_time"] = {
            "value": val,
            "unit": unit,
        }
        d["threads"] = {"value": self.merged_data.nb_threads, "unit": None}
        d["arch"] = {"value": platform.system(), "unit": None}
        d["python"] = {"value": platform.python_version(), "unit": None}
        d["track_types"] = {"value": self.merged_data.track_types, "unit": None}
        return d

    def __str__(self):
        s = ""
        for k, v in self.get_processed_output().items():
            if k == "track_types":
                if len(v["value"]) > 0:
                    s += "track_types\n"
                    for t, n in v["value"].items():
                        s += f"{' ' * 24}{t}: {n}\n"
            else:
                if v["unit"] is None:
                    unit = ""
                else:
                    unit = str(v["unit"])
                s += f"{k}{' ' * (20 - len(k))}{v['value']} {unit}\n"
        # remove last line break
        return s.rstrip("\n")

    def write_data(self, **kwargs):
        """Override virtual method from base class."""
        with open(self.get_output_path(which="merged"), "w+") as f:
            if self.encoder == "json":
                dump_json(self.get_processed_output(), f, indent=4)
            else:
                f.write(self.__str__())

    def write_data_if_requested(self, **kwargs):
        if self.write_to_disk is True:
            self.write_data(**kwargs)


class SimulationStatisticsActor(ActorBase, g4.GateSimulationStatisticsActor):
    """Store statistics about a simulation run."""

    # hints for IDE
    track_types_flag: bool

    user_info_defaults = {
        "track_types_flag": (
            False,
            {
                "doc": "Should the type of tracks be counted?",
            },
        ),
    }

    user_output_config = {
        "stats": {
            "actor_output_class": ActorOutputStatisticsActor,
        },
    }

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        # self._add_user_output(ActorOutputStatisticsActor, "stats")
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateSimulationStatisticsActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def __str__(self):
        s = self.user_output["stats"].__str__()
        return s

    @property
    def counts(self):
        return self.user_output.stats.merged_data

    def store_output_data(self, output_name, run_index, *data):
        raise NotImplementedError

    def initialize(self):
        ActorBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        g4.GateSimulationStatisticsActor.StartSimulationAction(self)
        self.user_output.stats.merged_data.nb_threads = (
            self.simulation.number_of_threads
        )

    def EndSimulationAction(self):
        g4.GateSimulationStatisticsActor.EndSimulationAction(self)
        self.user_output.stats.store_data(self.GetCounts())

        if self.simulation is not None:
            sim_start = self.simulation.run_timing_intervals[0][0]
        else:
            sim_start = 0

        if self.simulation is not None:
            sim_stop = self.simulation.run_timing_intervals[-1][1]
        else:
            sim_stop = 0

        self.user_output.stats.store_data(
            {"sim_start": sim_start, "sim_stop": sim_stop}
        )
        self.user_output.stats.merged_data.sim_start_time = (
            self.simulation.run_timing_intervals[0][0]
        )
        self.user_output.stats.merged_data.sim_stop_time = (
            self.simulation.run_timing_intervals[-1][1]
        )
        self.user_output.stats.merged_data.nb_threads = (
            self.simulation.number_of_threads
        )
        self.user_output.stats.write_data_if_requested()


class ActorOutputKillAccordingProcessesActor(ActorOutputBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.number_of_killed_particles = 0

    def get_processed_output(self):
        d = {}
        d["particles killed"] = self.number_of_killed_particles
        return d

    def __str__(self):
        s = ""
        for k, v in self.get_processed_output().items():
            s = k + ": " + str(v)
            s += "\n"
        return s


class KillAccordingProcessesActor(ActorBase, g4.GateKillAccordingProcessesActor):
    # hints for IDE
    processes_to_kill: list
    is_rayleigh_an_interaction: bool

    """
    This actor enables the user to kill particles according to one or more processes which occur in a volume. If the user
    wants to kill a particle whenever a proces occurs (except transportation), the "all" option is available.
    """

    user_info_defaults = {
        "processes_to_kill": (
            [],
            {
                "doc": "If a processes belonging to this list occured, the particle and its potential secondaries are killed. the variable all can be set up to kill a particle if an interaction occured."
            },
        ),
        "is_rayleigh_an_interaction": (
            True,
            {
                "doc": "Specific case to be faster. If a user wants to kill all interactions which implies an energy loss, this boolean enables to not account Rayleigh process as an interaction"
            },
        ),
    }
    user_output_config = {
        "kill_according_processes": {
            "actor_output_class": ActorOutputKillAccordingProcessesActor,
        },
    }

    """
    If a particle, not generated or generated within the volume at which our actor is attached, crosses the volume
    without interaction, the particle is killed.
    """

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()
        self.number_of_killed_particles = 0

    def __initcpp__(self):
        g4.GateKillAccordingProcessesActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunAction",
                "BeginOfEventAction",
                "PreUserTrackingAction",
                "SteppingAction",
                "EndSimulationAction",
            }
        )

    def initialize(self):
        ActorBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()
        if len(self.user_info.processes_to_kill) == 0:
            fatal("You have to select at least one process ! ")

    def EndSimulationAction(self):
        self.user_output.kill_according_processes.number_of_killed_particles = (
            self.number_of_killed_particles
        )

    def __str__(self):
        s = self.user_output["kill_according_processes"].__str__()
        return s


class KillActor(ActorBase, g4.GateKillActor):
    """Actor which kills a particle entering a volume."""

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        self.number_of_killed_particles = 0
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateKillActor.__init__(self, self.user_info)
        self.AddActions(
            {"StartSimulationAction", "EndSimulationAction", "SteppingAction"}
        )

    def initialize(self):
        ActorBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def EndSimulationAction(self):
        self.number_of_killed_particles = self.GetNumberOfKilledParticles()


class AttenuationImageActor(ActorBase, g4.GateAttenuationImageActor):
    """
    This actor generates an attenuation image for a simulation run.
    The output is a single image volume in cm^-1

    - image_volume: Input volume from which the attenuation map is generated.
    - energy: The energy level for which to generate the attenuation image.
    - database: The database source for attenuation coefficients, either 'EPDL' or 'NIST'.
    """

    # IDE hints
    image_volume = str
    energy = float
    database = str

    user_info_defaults = {
        "image_volume": (  # FIXME name or not name
            None,
            {
                "doc": "Input ImageVolume for which the attenuation map is generated.",
            },
        ),
        "energy": (
            None,
            {"doc": "The energy level for which to generate the attenuation image"},
        ),
        "database": (
            "EPDL",
            {
                "doc": "The database source for attenuation coefficients, either 'EPDL' or 'NIST'",
                "allowed_values": ("EPDL", "NIST"),
            },
        ),
    }

    user_output_config = {
        "attenuation_image": {
            "actor_output_class": ActorOutputSingleImage,
            "active": True,
            "write_to_disk": True,
            "keep_data_in_memory": True,
            "keep_data_per_run": True,
        },
    }

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateAttenuationImageActor.__init__(self, self.user_info)
        self.AddActions({"BeginOfRunAction"})

    def initialize(self):
        ActorBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def BeginOfRunAction(self, run):
        # the attenuation image is created during the first run only
        if run.GetRunID() != 0:
            return
        mu_image = self.image_volume.create_attenuation_image(
            self.database, self.energy
        )
        self.user_output.attenuation_image.store_data("merged", mu_image)
        self.user_output.attenuation_image.end_of_simulation()


process_cls(ActorOutputStatisticsActor)
process_cls(SimulationStatisticsActor)
process_cls(KillActor)
process_cls(ActorOutputKillAccordingProcessesActor)
process_cls(KillAccordingProcessesActor)
process_cls(AttenuationImageActor)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  2 09:48:30 2025

@author: fava
"""
import opengate as gate
from opengate.utility import g4_units
import opengate.tests.utility as utility
import uproot
import numpy as np
from scipy.spatial.transform import Rotation


if __name__ == "__main__":

    paths = utility.get_default_test_paths(__file__, output_folder="gate_test090")

    # units
    cm = g4_units.cm
    mm = g4_units.mm
    nm = g4_units.nm
    keV = gate.g4_units.keV

    # create simulation object
    sim = gate.Simulation()
    sim.output_dir = paths.output
    sim.g4_verbose = False
    sim.visu = False
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    sim.number_of_threads = 1
    sim.random_seed = 321654
    sim.output_dir = paths.output

    # world
    sim.world.size = [100 * cm, 100 * cm, 100 * cm]
    sim.world.material = "G4_Galactic"

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    plane.mother = sim.world
    plane.material = "G4_Galactic"
    plane.rmin = 0
    plane.rmax = 7 * cm
    plane.dz = 1 * nm  # half height
    rot = Rotation.from_rotvec(np.pi / 2 * np.array([0, 1, 0]))
    plane.rotation = rot.as_matrix()
    plane.translation = [
        30 * cm,
        0,
        0,
    ]
    plane.color = [1, 0, 0, 1]  # red

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = sim.world
    waterbox.material = "G4_WATER"
    waterbox.size = [2 * cm, 2 * cm, 2 * cm]
    waterbox.translation = [0, 0, 0 * mm]
    waterbox.color = [0, 0, 1, 1]  # blue

    # physics
    sim.physics_manager.physics_list_name = "G4EmLivermorePolarizedPhysics"

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.particle = "gamma"
    source.energy.type = "gauss"
    source.energy.mono = 100 * keV
    source.position.type = "disc"
    source.position.radius = 1 * mm
    source.position.translation = [0, 0, -30 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 1e6

    # PhaseSpace Actor to save polarization
    phspa = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phspa.attached_to = plane.name
    phspa.attributes = [
        "Polarization",
    ]
    phspa.output_filename = "test091_phsp_actor_no_defined_polarization.root"

    # Run the simulation without user defined polarization
    print("Run the simulation without user defined polarization")
    sim.run(start_new_process=True)

    # The polarization outputs should be random
    root_filename = paths.output / "test091_phsp_actor_no_defined_polarization.root"
    root_file = uproot.open(root_filename)["PhaseSpace"]
    px = root_file["Polarization_X"].array(library="np")
    nb_compton_no_polarization = px.size
    print()

    # Set the polarization to a specific value
    source.polarization = [1, 0, 0]  # linear polarization (horizontal)
    # source.polarization = [-1, 0, 0]  # linear polarization (vertical)
    # source.polarization = [0, 1, 0]  # linear polarization (45°)
    # source.polarization = [0, -1, 0]  # linear polarization (-45°)
    # source.polarization = [0, 0, 1]  # circular polarization (right)
    # source.polarization = [0, 0, -1]  # circular polarization (left)
    # source.polarization = [0, 0, 0]  # unpolarized
    phspa.output_filename = "test091_phsp_actor_with_polarization.root"
    print("Run the simulation with user defined linear polarization")
    sim.run(start_new_process=True)

    # The polarization outputs should be 1, 0, 0
    root_filename = paths.output / "test091_phsp_actor_with_polarization.root"
    root_file = uproot.open(root_filename)["PhaseSpace"]
    px = root_file["Polarization_X"].array(library="np")
    nb_compton_with_polarization = px.size

    # Compute std (poisson distribution)
    std = np.sqrt(nb_compton_no_polarization)
    is_ok = nb_compton_with_polarization < (nb_compton_no_polarization - 5 * std)
    utility.print_test(
        is_ok,
        "Compton detected without polarization: "
        + str(nb_compton_no_polarization)
        + " > "
        + "Compton detected with polarization: "
        + str(nb_compton_with_polarization),
    )
    utility.test_ok(is_ok)



import tables
import numpy as np

from opengate.postprocessors.base import PostProcessor
from opengate import Simulation

__run_group_prefix__ = "data_from_run"


def create_mock_data(path):
    n_rows = int(1e5)

    class Attributes(tables.IsDescription):
        event_id = tables.FloatCol(pos=1)
        particle = tables.StringCol(itemsize=16, pos=2)
        energy = tables.FloatCol(pos=3)
        position = tables.FloatCol(shape=(3,), pos=4)

    # dt = np.dtype([('event_id', np.float32),
    #                ('particle', 'S20'),
    #                ('energy', np.float32),
    #                ('position', np.float32),
    #                ])

    with tables.open_file(path, mode="w", title="Test file") as h5file:
        runs = list(range(10))
        for g_name in [f"{__run_group_prefix__}_{r}" for r in runs] + ["merged_data"]:
            group = h5file.create_group("/", g_name, f"Data from {g_name}")
            table = h5file.create_table(group, "singles", description=Attributes)
            table.cols.energy.attrs["unit"] = "MeV"

            arr_event_id = np.arange(n_rows)
            arr_energy = np.ones(n_rows).reshape(n_rows) * 300
            arr_particles = np.array(["proton"] * n_rows)

            x = np.random.randn(n_rows) * 20
            y = np.random.randn(n_rows) * 20
            z = np.ones_like(x)
            arr_position = np.vstack((x, y, z)).T

            table.append([arr_event_id, arr_particles, arr_energy, arr_position])


path = "postprocessor_test.h5"
create_mock_data(path)

sim = Simulation()
sim.output_path = "."

post_processor = PostProcessor(name="post_processor", simulation=sim)

datafetcher = post_processor.add_processing_unit(
    "DataFetcherHdf5",
    name="datafetcher",
    input_path="postprocessor_test.h5",
    hdf5_node_name="singles",
    input_hdf5_group="/merged_data",
)

# sequence = post_processor.add_processing_unit('ProcessingSequence', name='sequence1')
#
# gaussian_blurrer_energy = sequence.add_processing_unit('GaussianBlurringSingleAttribute',
#                                                        name='blurrer_energy',
#                                                        attribute='energy',
#                                                        sigma=10)
#
# gaussian_blurrer_position = sequence.add_processing_unit('GaussianBlurringSingleAttribute',
#                                                          name='blurrer_position',
#                                                          attribute='position',
#                                                          sigma=10)
#
# offsetter_energy = sequence.add_processing_unit('OffsetSingleAttribute',
#                                                 name='offset_energy',
#                                                 attribute='energy',
#                                                 offset=10)

projector = post_processor.add_processing_unit(
    "ProjectionListMode", name="projector", size=[100, 100, 1], spacing=[1.2, 0.9, 1]
)


#
# # gaussian_blurrer_energy = GaussianBlurringSingleAttribute(name='blurrer_energy',
# #                                                           attribute='energy',
# #                                                           sigma=10)
# #
# # gaussian_blurrer_position = GaussianBlurringSingleAttribute(name='blurrer_position',
# #                                                             attribute='position',
# #                                                             sigma=10)
# #
# # offsetter_energy = OffsetSingleAttribute(name='offset_energy',
# #                                          attribute='energy',
# #                                          offset=10)
# # datafetcher = DataFetcherHdf5(name='datafetcher',
# #                               input_path="postprocessor_test.h5",
# #                               hdf5_node_name='singles',
# #                               path_in_hdf5_file='/merged_data')
# #
# #
# # sequence1 = ProcessingSequence(name='sequence1')
#
# sequence1.add_processing_unit(gaussian_blurrer_energy)
# sequence1.add_processing_unit(gaussian_blurrer_position)
# sequence1.add_processing_unit(offsetter_energy)


# post_processor.add_processing_unit(datafetcher)
# post_processor.add_processing_unit(sequence1)

# post_processor.initialize()

post_processor.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import opengate as gate
from opengate.tests import utility
from opengate.tests.src.test081_tle_helpers import (
    add_waterbox,
    add_source,
    plot_pdd,
    compare_pdd,
)
import sys


def main(argv):
    paths = utility.get_default_test_paths(__file__, output_folder="test081_tle")

    # create the simulation
    sim = gate.Simulation()

    # main options
    # sim.visu = True
    sim.visu_type = "qt"
    sim.random_seed = 12356654
    sim.output_dir = paths.output
    sim.progress_bar = True
    sim.number_of_threads = 1
    if len(argv) > 1:
        sim.number_of_threads = int(argv[1])

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # create voxelized waterbox
    waterbox = add_waterbox(sim)

    # save voxelized versions of the waterbox
    # voxelize_waterbox(sim, paths.data / "test081_tle")

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.global_production_cuts.all = 1 * mm
    sim.physics_manager.set_max_step_size("waterbox", 1 * mm)
    sim.physics_manager.set_user_limits_particles("gamma")

    s = f"/process/eLoss/CSDARange true"
    sim.g4_commands_before_init.append(s)

    # default source for tests
    source = add_source(sim, n=2e5)

    tle_dose_actor = sim.add_actor("TLEDoseActor", "tle_dose_actor")
    tle_dose_actor.output_filename = "test081_tle.mhd"
    tle_dose_actor.attached_to = waterbox
    tle_dose_actor.dose_uncertainty.active = True
    tle_dose_actor.dose.active = True
    tle_dose_actor.size = [200, 200, 200]
    tle_dose_actor.spacing = [x / y for x, y in zip(waterbox.size, tle_dose_actor.size)]
    print(f"TLE Dose actor pixels : {tle_dose_actor.size}")
    print(f"TLE Dose actor spacing : {tle_dose_actor.spacing} mm")
    print(f"TLE Dose actor size : {waterbox.size} mm")

    # add conventional dose actor
    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    dose_actor.output_filename = "test081.mhd"
    dose_actor.attached_to = waterbox
    dose_actor.dose_uncertainty.active = True
    dose_actor.dose.active = True
    dose_actor.size = [200, 200, 200]
    dose_actor.spacing = [x / y for x, y in zip(waterbox.size, dose_actor.size)]
    print(f"Dose actor pixels : {dose_actor.size}")
    print(f"Dose actor spacing : {dose_actor.spacing} mm")
    print(f"Dose actor size : {waterbox.size} mm")

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # print results at the end
    print(stats)
    ax, plt = plot_pdd(dose_actor, tle_dose_actor)
    f1 = dose_actor.edep.get_output_path()
    f2 = tle_dose_actor.edep.get_output_path()
    is_ok = compare_pdd(f1, f2, dose_actor.spacing[2], ax[0], tol=0.25)

    print()
    f1 = dose_actor.dose.get_output_path()
    f2 = tle_dose_actor.dose.get_output_path()
    is_ok = compare_pdd(f1, f2, dose_actor.spacing[2], ax[1], tol=0.25) and is_ok

    # output
    f = paths.output / f"pdd_geom.png"
    plt.savefig(f)
    print(f"PDD image saved in {f}")

    utility.test_ok(is_ok)


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import gatetools as gt
import itk
import numpy as np

import opengate as gate
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test015")

    # create the simulation
    sim = gate.Simulation()

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm

    # main options
    sim.visu = False
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = True
    sim.random_seed = 123654789

    # world size
    sim.world.size = [0.5 * m, 0.5 * m, 0.5 * m]

    # add an iec phantom
    iec_phantom = gate_iec.add_iec_phantom(sim)
    iec_phantom.translation = [0 * cm, 0 * cm, 0 * cm]

    # output filename
    f = paths.output / "test015_iec_1.mhd"

    # # voxelize the iec
    labels, image = sim.voxelize_geometry(
        extent=iec_phantom, spacing=[3, 3, 3], margin=1
    )
    print(f"Labels : ")
    for k, v in labels.items():
        print(f"{k} = {v}")
    # rotate the IEC to be like the reference CT
    # 1) rotation 180 around X to be like in the iec 61217 coordinate system
    # 2) rotation 180 around Y because we put the phantom in that orientation on the table
    # (in reality there is an additional tiny rotation around Z, maybe 3 deg, but we don't care here)
    image = gt.applyTransformation(
        input=image, force_resample=True, adaptive=True, rotation=(180, 180, 0)
    )
    # the translation is computed as follows:
    # ref point in ref image  : 10 -30 477
    # ref point in test image : 3 153 27
    rp_ri = np.array([10, -30, 477])
    rp_ti = np.array([3, 153, 27])
    t = rp_ri - rp_ti
    t = np.array(list(image.GetOrigin())) + t
    image.SetOrigin(t)
    print(f"Write image {f}")
    itk.imwrite(image, str(f))

    # compare image
    print("Image can be compared with : ")
    gate.exception.warning(f"vv {paths.output_ref / 'iec_ct_3mm.mhd '} --fusion {f}")

    # Comparison with reference
    rf = paths.output_ref / "test015_iec_1.mhd"
    print(f"Reference image : {rf}")
    print(f"Computed  image : {f}")
    is_ok = gate.image.compare_itk_image(rf, f)
    utility.print_test(is_ok, "Compare images")
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, None, "test054")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.visu_type = "qt"
    # sim.visu = True
    sim.check_volumes_overlap = True
    sim.number_of_threads = 1
    sim.random_seed = 654923
    sim.output_dir = paths.output

    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm
    sec = gate.g4_units.second

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_AIR"

    # image
    patient = sim.add_volume("Box", "patient")
    patient.material = "G4_WATER"
    patient.size = [200 * mm, 200 * mm, 200 * mm]
    patient.translation = [0, 0, 0]

    # create two other parallel worlds
    sim.add_parallel_world("world2")

    # detector in w2 (on top of world)
    det = sim.add_volume("Box", "detector")
    det.mother = "world2"
    det.material = "NaI"
    det.size = [400 * mm, 400 * mm, 40 * mm]
    det.translation = [0, 0, 100 * mm]

    n = 2
    interval_length = 1 * sec / n
    sim.run_timing_intervals = [
        (i * interval_length, (i + 1) * interval_length) for i in range(n)
    ]
    gantry_angles_deg = [i * 20 for i in range(n)]
    (
        dynamic_translations,
        dynamic_rotations,
    ) = gate.geometry.utility.get_transform_orbiting(
        initial_position=det.translation, axis="Y", angle_deg=gantry_angles_deg
    )
    print(dynamic_translations)
    print(dynamic_rotations)
    print(sim.run_timing_intervals)

    det.add_dynamic_parametrisation(
        rotation=dynamic_rotations, translation=dynamic_translations
    )

    # source
    source = sim.add_source("GenericSource", "source")
    source.energy.mono = 0.144 * MeV
    source.particle = "gamma"
    source.position.type = "box"
    source.position.size = [1 * mm, 200 * mm, 1 * mm]
    source.position.translation = [0, 0, 0 * cm]
    source.activity = 1e5 * Bq / sim.number_of_threads
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.attached_to = det
    hc.attributes = [
        "KineticEnergy",
        "PostPosition",
        "PrePosition",
        "TotalEnergyDeposit",
        "GlobalTime",
        "TrackVolumeName",
        "TrackID",
        "PostStepUniqueVolumeID",
        "PreStepUniqueVolumeID",
        "TrackVolumeCopyNo",
        "TrackVolumeInstanceID",
        "TrackVertexKineticEnergy",
        "TrackVertexPosition",
        "ProcessDefinedStep",
        "EventPosition",
    ]
    hc.output_filename = "hits.root"

    cc = sim.add_actor("DigitizerEnergyWindowsActor", "EnergyWindows")
    keV = gate.g4_units.keV
    channels = [
        # {"name": f"spectrum_{detector}", "min": 3 * keV, "max": 515 * keV},
        {"name": f"scatter", "min": 108.57749938965 * keV, "max": 129.5924987793 * keV},
        {"name": f"peak140", "min": 129.5924987793 * keV, "max": 150.60751342773 * keV},
    ]
    cc.attached_to = hc.attached_to
    cc.input_digi_collection = hc.name
    cc.channels = channels
    cc.output_filename = hc.output_filename

    # projection
    proj = sim.add_actor("DigitizerProjectionActor", "Projection")
    proj.attached_to = cc.attached_to
    proj.input_digi_collections = [x["name"] for x in cc.channels]

    # Only record primary window
    proj.input_digi_collections = ["peak140"]
    proj.spacing = [2.46 * mm, 2.46 * mm]
    proj.size = [128, 16]
    proj.origin_as_image_center = True
    proj.output_filename = "projections.mha"

    # --------------------------------------------------------------------------
    print("Geometry trees: ")
    print(sim.volume_manager.dump_volume_tree())

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # check image
    is_ok = utility.assert_images(
        paths.output_ref / "projections.mha",
        proj.get_output_path("counts"),
        stats=stats,
        tolerance=40,
        axis="x",
        sad_profile_tolerance=15,
        fig_name=sim.output_dir / "parallel_proj.png",
    )
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, output_folder="test047_gan_vox_source"
    )

    """
    Data :
    gt_image_crop source_three_areas.mhd -o source_three_areas_crop.mhd
    gt_affine_transform -i source_three_areas_crop.mhd -o source_three_areas_crop_3.5mm.mhd  --newspacing 3.5 -fr -a -im NN
    """

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.number_of_threads = 1
    sim.random_seed = 123456789
    sim.output_dir = paths.output
    sim.progress_bar = True
    activity_bq = 1e6

    # visu
    if sim.visu:
        sim.number_of_threads = 1
        activity_bq = 100

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    keV = gate.g4_units.keV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    gcm3 = gate.g4_units.g_cm3

    #  change world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # add a fake object to test position & rotation
    fake = sim.add_volume("Box", "fake")
    fake.size = [1 * m, 1 * m, 1 * m]
    fake.translation = [5 * cm, 10 * cm, 0 * cm]
    r = Rotation.from_euler("y", 33, degrees=True)
    fake.rotation = r.as_matrix()
    fake.material = "G4_AIR"
    fake.color = [0, 0, 1, 1]  # blue

    # add an image
    f = paths.data / "ct_4mm.mhd"
    if sim.visu:
        f = paths.data / "ct_40mm.mhd"

    ct = sim.add_volume("Image", "ct")
    ct.image = f
    ct.mother = "fake"
    ct.material = "G4_AIR"
    tol = 0.1 * gcm3
    ct.voxel_materials, materials = gate.geometry.materials.HounsfieldUnit_to_material(
        sim,
        tol,
        paths.data / "Schneider2000MaterialsTable.txt",
        paths.data / "Schneider2000DensitiesTable.txt",
    )
    ct.dump_label_image = paths.output / "ct_4mm_labels.mhd"

    # condGAN source with voxelized condition
    source = sim.add_source("GANSource", "source")
    source.attached_to = "ct"
    source.cond_image = paths.data / "source_three_areas_crop_3.5mm.mhd"
    source.position.translation = gate.image.get_translation_between_images_center(
        str(ct.image), str(source.cond_image)
    )
    source.position.translation = source.position.translation / 2.0
    print(f"translation {source.position.translation}")
    source.particle = "alpha"
    source.activity = activity_bq * Bq / sim.number_of_threads
    source.compute_directions = True
    source.pth_filename = paths.data / "train_gaga_v001_GP_0GP_10_60000.pth"
    source.position_keys = ["PrePosition_X", "PrePosition_Y", "PrePosition_Z"]
    # source.backward_distance = 20 * cm
    # source.backward_force = True
    source.cond_debug = True
    source.direction_keys = ["PreDirection_X", "PreDirection_Y", "PreDirection_Z"]
    source.energy.mono = 1 * MeV
    source.energy_key = None
    source.weight_key = None
    source.time_key = None
    source.relative_timing = True
    source.batch_size = 1e5
    source.verbose_generator = True
    source.gpu_mode = (
        utility.get_gpu_mode_for_tests()
    )  # should be "auto" but "cpu" for macOS github actions to avoid mps errors

    # cuts (not need precision here)
    c = sim.physics_manager.global_production_cuts.all = 100 * mm

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test047.mhd"
    dose.attached_to = "ct"
    dose.size = [70, 70, 240]
    dose.spacing = [3.5 * mm, 3.5 * mm, 3.5 * mm]
    dose.output_coordinate_system = "attached_to_image"
    dose.hit_type = "random"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # ---------------------------------------------------------------
    # print results at the end
    print()
    gate.exception.warning("Tests stats file")
    print(stats)
    ref_stat_file = paths.output_ref / "t047_stats.txt"
    # stat.write(ref_stat_file) # (for reference)
    stats_ref = utility.read_stat_file(ref_stat_file)
    is_ok = utility.assert_stats(stats, stats_ref, 0.005)

    dose = sim.get_actor("dose")
    print()
    gate.exception.warning("Compare image to analog")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test047-edep.mhd",
            dose.get_output_path("edep"),
            stats,
            tolerance=19,
            ignore_value_data2=0,
            axis="x",
        )
        and is_ok
    )

    print("Test with vv: ")
    print(f"vv {source.cond_image} --fusion {dose.get_output_path('edep')}")

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test090_helpers import *

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test090_spect_config_1"
    )

    # TEST 1: intevo, Lu177, 2 heads, 3 angles

    # delete the output path content
    utility.delete_folder_contents(paths.output)
    data_path = paths.data
    mm = gate.g4_units.mm

    sc = SPECTConfig()
    sc.simu_name = "test090"
    sc.output_folder = paths.output
    sc.number_of_threads = 1
    # detectors
    sc.detector_config.model = "intevo"
    sc.detector_config.collimator = "melp"
    sc.detector_config.number_of_heads = 2
    sc.detector_config.digitizer_function = intevo.add_digitizer
    sc.detector_config.size = [256, 256]
    sc.detector_config.spacing = [2.39759994 * mm, 2.39759994 * mm]
    # phantom
    sc.phantom_config.image = data_path / "iec_5mm.mhd"
    sc.phantom_config.labels = data_path / "iec_5mm_labels.json"
    sc.phantom_config.material_db = data_path / "iec_5mm.db"
    # source
    sc.source_config.image = data_path / "iec_5mm_activity.mhd"
    sc.source_config.radionuclide = "177lu"
    sc.source_config.total_activity = 5e4 * gate.g4_units.Bq
    # acquisition
    sc.acquisition_config.radius = 300 * gate.g4_units.mm
    sc.acquisition_config.duration = 30 * gate.g4_units.s
    sc.acquisition_config.number_of_angles = 3

    # create the simulation
    print(sc)
    sim = gate.Simulation()
    sc.setup_simulation(sim, visu=False)
    stats = sim.actor_manager.find_actors("stats")[0]

    # run it
    sim.random_seed = 987654
    sim.run()

    # we check only that the output files exist
    is_ok = True
    is_ok = check_stats_file(272198, sc, stats, is_ok)
    is_ok = check_projection_files(sim, paths, stats, is_ok)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pickle

import opengate as gate
import opengate.tests.utility as tu


if __name__ == "__main__":
    # create the simulation
    sim = gate.Simulation()
    sim.verbose_getstate = True

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")

    # add actor1
    actor1 = sim.add_actor("DoseActor", "actor1")
    actor1.attached_to = "waterbox"
    actor1.output_filename = "original.mhd"
    actor1.edep_uncertainty.active = True
    actor1.edep.keep_data_per_run = True

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True

    kill = sim.add_actor("KillActor", "kill")

    print("****************  PICKLE SIM  ****************")
    sim_pickled = pickle.dumps(sim)
    print("***************  UNPICKLE SIM ***************")
    sim_reloaded = pickle.loads(sim_pickled)

    stat_reloaded = sim_reloaded.get_actor("Stats")
    actor1_reloaded = sim_reloaded.get_actor("actor1")
    actor1_reloaded.edep.output_filename = "reloaded.mhd"
    actor1_reloaded.edep_uncertainty.active = False

    sim_pickled_twice = pickle.loads(pickle.dumps(sim_reloaded))

    actor1_reloaded_twice = sim_pickled_twice.get_actor("actor1")

    is_ok = True
    print("**** TEST ****")

    b = actor1_reloaded_twice.edep.output_filename != actor1.edep.output_filename
    is_ok &= b
    print(
        f"Test: Is the input parameter 'output_filename' equal in original and "
        f"re-re-loaded actor output to which interface edep belongs? \n"
        f"Expected: not equal. Passed: {b}"
    )

    b = (
        actor1_reloaded_twice.edep_uncertainty.active
        is not actor1.edep_uncertainty.active
    )
    is_ok &= b
    print(
        f"Test: Is the input parameter 'active' equal in original and "
        f"re-re-loaded actor output to which interface edep_uncertainty belongs? \n"
        f"Expected: not equal. Passed: {b}"
    )

    b = stat.stats.belongs_to_actor is stat
    is_ok &= b
    print(f"interface in original actor belongs to original actor? Passed: {b}")

    b = stat_reloaded.stats.belongs_to_actor is stat_reloaded
    is_ok &= b
    print(f"interface in reloaded belongs to reloaded actor? Passed: {b}")

    tu.test_ok(is_ok)



#!/usr/bin/env python3

import site
import os
import click
import pkgutil
from pathlib import Path


def return_site_packages_dir() -> Path:
    site_package = [p for p in site.getsitepackages() if "site-packages" in p][0]
    return Path(site_package)


def get_site_packages_dir():
    print(str(return_site_packages_dir()))


def get_libG4processes_path():
    lib_path = return_site_packages_dir() / "opengate_core.libs"
    for element in lib_path.iterdir():
        if "libG4processes" in element.name:
            print(str(element))


def get_libG4geometry_path():
    lib_path = return_site_packages_dir() / "opengate_core.libs"
    for element in lib_path.iterdir():
        if "libG4geometry" in element.name:
            print(str(element))


def return_tests_path():
    pathFile = Path(__file__).parent.resolve()
    if "src" in pathFile.iterdir():
        mypath = pathFile.parent / "tests" / "src"
    else:
        mypath = (
            Path(pkgutil.get_loader("opengate").get_filename()).resolve().parent
            / "tests"
            / "src"
        )
    return mypath


def get_tests_path():
    print(str(return_tests_path()))


# -----------------------------------------------------------------------------
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("--path", "-p", default="", help="Path", required=True)
def go(path):
    """
    Tool to have the path of folders
    """
    if path == "libG4processes":
        get_libG4processes_path()
    elif path == "libG4geometry":
        get_libG4geometry_path()
    elif path == "site_packages":
        get_site_packages_dir()
    elif path == "tests":
        get_tests_path()


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
import numpy as np
import opengate.contrib.spect.ge_discovery_nm670 as gate_spect
import opengate as gate
import test043_garf_helpers as test43
from opengate.tests import utility


if __name__ == "__main__":
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 1236546
    sim.number_of_threads = 1
    colli = "lehr"

    # units
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    Bq = gate.g4_units.Bq
    keV = gate.g4_units.keV

    activity = 1e6 * Bq / sim.number_of_threads

    # world
    test43.sim_set_world(sim)

    # spect head
    spect, colli, crystal = gate_spect.add_spect_head(
        sim, "spect", collimator_type=colli, debug=sim.visu
    )
    spect.translation = [0, 0, -15 * cm]
    crystal_name = f"{spect.name}_crystal"

    # physics
    test43.sim_phys(sim)

    # sources
    test43.sim_source_test(sim, activity)

    # digitizer
    channels = [
        {"name": f"spectrum_{spect.name}", "min": 0 * keV, "max": 160 * keV},
        {"name": f"scatter_{spect.name}", "min": 114 * keV, "max": 126 * keV},
        {"name": f"peak140_{spect.name}", "min": 126 * keV, "max": 154 * keV},
    ]
    proj = gate_spect.add_digitizer_OLD(sim, crystal_name, channels)
    proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.output_filename = test43.paths.output / "test043_projection_analog.mhd"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # dump the output image with offset like in old gate (for comparison)
    print("We change the spacing/origin to be compared to the old gate")
    spacing = np.array([4.41806 * mm, 4.41806 * mm, 1])
    img = itk.imread(str(proj.get_output_path("counts")))
    img.SetSpacing(spacing)
    img.SetOrigin(spacing / 2.0)
    fn = str(proj.get_output_path("counts")).replace(".mhd", "_offset.mhd")
    itk.imwrite(img, fn)

    # ----------------------------------------------------------------------------------------------------------------
    # tests
    print()
    gate.exception.warning("Tests stats file")
    stats_ref = utility.read_stat_file(test43.paths.gate_output / "stats_analog.txt")
    stats.counts.runs = 1  # force to one run (ref only have 1 thread)
    is_ok = utility.assert_stats(stats, stats_ref, 0.01)

    print()
    gate.exception.warning("Tests projection (old gate)")
    is_ok = (
        utility.assert_images(
            test43.paths.gate_output / "projection_analog.mhd",
            fn,
            stats,
            tolerance=75,
            ignore_value_data2=0,
            axis="x",
            sum_tolerance=35,
        )
        and is_ok
    )

    print()
    gate.exception.warning("Tests projection (new)")
    is_ok = (
        utility.assert_images(
            test43.paths.output_ref / "test043_projection_analog.mhd",
            proj.get_output_path("counts"),
            stats,
            tolerance=80,
            ignore_value_data2=0,
            axis="x",
            sum_tolerance=25,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



import numpy as np

from .unitbase import ProcessingUnitBase
from ..exception import fatal
from .utility import get_table_column


class ListModeSingleAttribute(ProcessingUnitBase):
    _required_number_of_input_units = 1

    user_info_defaults = {
        "attribute": (
            0,
            {
                "doc": "Name of the attribute to be blurred.",
            },
        ),
    }

    def assert_input_is_compatible(self):
        if len(self.children) != self._required_number_of_input_units:
            fatal(f"Too many input units. This unit accepts only one. ")

    def initialize_output_data_handle(self):
        input_data_handle = self.input_units[0].output_data_handle
        self.output_data_handle = input_data_handle.copy(
            newparent=self.hdf5_group
        )  # , newname=get_node_name(input_data_handle))

    def get_table_column_to_work_on(self):
        if self.output_data_handle is None:
            self.initialize_output_data_handle()
        return get_table_column(self.output_data_handle, self.attribute)


class GaussianBlurringSingleAttribute(ListModeSingleAttribute):
    user_info_defaults = {
        "sigma": (
            0,
            {
                "doc": "1-sigma width of the blurring kernel.",
            },
        ),
    }

    def do_your_job(self):
        column = self.get_table_column_to_work_on()
        try:
            attribute_size = column.shape[1]
        except IndexError:
            attribute_size = 1
        try:
            sigma_size = len(self.sigma)
        except TypeError:
            sigma_size = 1
        if sigma_size == 1:
            sigma = np.array([self.sigma] * attribute_size)
            # perturbation = np.random.randn(len(column)) * self.sigma
        elif sigma_size == attribute_size:
            sigma = np.array([self.sigma])
        else:
            fatal(
                f"User inout sigma={self.sigma} incompatible with the attribute of length {attribute_size}."
            )

        perturbation = np.multiply(
            np.random.randn(len(column) * len(sigma)).reshape(len(column), len(sigma)),
            sigma,
        ).squeeze()
        column[:] += perturbation


class OffsetSingleAttribute(ListModeSingleAttribute):
    user_info_defaults = {
        "offset": (
            0,
            {
                "doc": "1-sigma width of the blurring kernel.",
            },
        ),
    }

    def do_your_job(self):
        column = self.get_table_column_to_work_on()  # this is a column object
        column[:] += self.offset  # with [:], we get an array
